# Copyright 2014 Rustici Software
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
.. module:: base
   :synopsis: A base object to provide a common initializer for objects to
   easily keep track of required and optional properties and their
   error checking

"""


class Base(object):
    _props = []

    def __init__(self, *args, **kwargs):
        """Initializes an object by checking the provided arguments
        against lists defined in the individual class. If required
        properties are defined, this method will set them to None by default.
        Optional properties will be ignored if they are not provided. The
        class may provide custom setters for properties, in which case those
        setters (see __setattr__ below).

        """
        if hasattr(self, '_props_req') and self._props_req:
            list(map(lambda k: setattr(self, k, None), self._props_req))

        new_kwargs = {}
        for obj in args:
            new_kwargs.update(obj if isinstance(obj, dict) else vars(obj))

        new_kwargs.update(kwargs)

        for key, value in new_kwargs.items():
            setattr(self, key, value)

    def __setattr__(self, attr, value):
        """Makes sure that only allowed properties are set. This method will
        call the proper attribute setter as defined in the class to provide
        additional error checking

        :param attr: the attribute being set
        :type attr: str
        :param value: the value to set

        """
        if attr.startswith('_') and attr[1:] in self._props:
            super(Base, self).__setattr__(attr, value)
        elif attr not in self._props:
            raise AttributeError(
                f"Property '{attr}' cannot be set on a 'tincan.{self.__class__.__name__}' object."
                f"Allowed properties: {', '.join(self._props)}"
            )
        else:
            super(Base, self).__setattr__(attr, value)

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__



# Copyright 2014 Rustici Software
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from tincan.serializable_base import SerializableBase


class Score(SerializableBase):
    """Stores the scoring data for an activity.

    Can be created from a dict, another Score, or from kwargs.

    All these attributes are optional and settable to None:

    :param scaled: Scaled score
    :type scaled: float
    :param raw: Raw score
    :type raw: float
    :param min: Minimum score possible
    :type min: float
    :param max: Maximum score possible
    :type max: float
    """

    _props = [
        'scaled',
        'raw',
        'min',
        'max',
    ]

    def __init__(self, *args, **kwargs):
        self._scaled = None
        self._raw = None
        self._min = None
        self._max = None

        super(SerializableBase, self).__init__(*args, **kwargs)

    @property
    def scaled(self):
        """Scaled for Score

        :setter: Tries to convert to float. If None is provided,
        this signifies the absence of this data.
        :setter type: float | int | None
        :rtype: float | None
        :raises: TypeError if unsupported type is provided
        """
        return self._scaled

    @scaled.setter
    def scaled(self, value):
        if value is None or isinstance(value, float):
            self._scaled = value
            return
        try:
            self._scaled = float(value)
        except Exception as e:
            msg = (
                f"Property 'scaled' in a 'tincan.{self.__class__.__name__}' object must be set with a "
                f"float or None."
            )
            msg += repr(e)
            raise TypeError(msg)

    @scaled.deleter
    def scaled(self):
        del self._scaled

    @property
    def raw(self):
        """Raw for Score

        :setter: Tries to convert to float. If None is provided,
        this signifies the absence of this data.
        :setter type: float | int | None
        :rtype: float | None
        :raises: TypeError if unsupported type is provided
        """
        return self._raw

    @raw.setter
    def raw(self, value):
        if value is None or isinstance(value, float):
            self._raw = value
            return
        try:
            self._raw = float(value)
        except Exception as e:
            msg = (
                f"Property 'raw' in a 'tincan.{self.__class__.__name__}' object must be set with a "
                f"float or None."
            )
            msg += repr(e)
            raise TypeError(msg)

    @raw.deleter
    def raw(self):
        del self._raw

    @property
    def min(self):
        """Min for Score

        :setter: Tries to convert to float. If None is provided,
        this signifies the absence of this data.
        :setter type: float | int | None
        :rtype: float | None
        :raises: TypeError if unsupported type is provided
        """
        return self._min

    @min.setter
    def min(self, value):
        if value is None or isinstance(value, float):
            self._min = value
            return
        try:
            self._min = float(value)
        except Exception as e:
            msg = (
                f"Property 'min' in a 'tincan.{self.__class__.__name__}' object must be set with a "
                f"float or None."
            )
            msg += repr(e)
            raise TypeError(msg)

    @min.deleter
    def min(self):
        del self._min

    @property
    def max(self):
        """Max for Score

        :setter: Tries to convert to float. If None is provided,
        this signifies the absence of this data.
        :setter type: float | int | None
        :rtype: float | None
        :raises: TypeError if unsupported type is provided
        """
        return self._max

    @max.setter
    def max(self, value):
        if value is None or isinstance(value, float):
            self._max = value
            return
        try:
            self._max = float(value)
        except Exception as e:
            msg = (
                f"Property 'max' in a 'tincan.{self.__class__.__name__}' object must be set with a "
                f"float or None."
            )
            msg += repr(e)
            raise TypeError(msg)

    @max.deleter
    def max(self):
        del self._max



# Copyright 2014 Rustici Software
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from http.client import HTTPResponse

from tincan.http_request import HTTPRequest
from tincan.base import Base


class LRSResponse(Base):
    """Creates a new LRSResponse object, either from a dict, another object, or from kwargs

    :param success: True if the LRS return a successful status (sometimes includes 404), False otherwise
    :type success: bool
    :param request: HTTPRequest object that was sent to the LRS
    :type request: HTTPRequest
    :param response: HTTPResponse object that was received from the LRS
    :type response: HTTPResponse
    :param data: Body of the HTTPResponse
    :type data: unicode
    :param content: Parsed content received from the LRS
    """

    _props_req = [
        'success',
        'request',
        'response',
        'data',
    ]

    _props = [
        'content',
    ]

    _props.extend(_props_req)

    def __init__(self, *args, **kwargs):
        self._success = None
        self._request = None
        self._response = None
        self._data = None
        self._content = None

        super(LRSResponse, self).__init__(*args, **kwargs)

    @property
    def success(self):
        """The LRSResponse's success. True if the LRS return a
        successful status (sometimes includes 404), False otherwise.

        :setter: Tries to convert to boolean
        :setter type: bool
        :rtype: bool
        """
        return self._success

    @success.setter
    def success(self, value):
        self._success = bool(value)

    @property
    def request(self):
        """The HTTPRequest object that was sent to the LRS

        :setter: Tries to convert to an HTTPRequest object
        :setter type: :class:`tincan.http_request.HTTPRequest`
        :rtype: :class:`tincan.http_request.HTTPRequest`
        """
        return self._request

    @request.setter
    def request(self, value):
        if value is not None and not isinstance(value, HTTPRequest):
            value = HTTPRequest(value)

        self._request = value

    @property
    def response(self):
        """The HTTPResponse object that was sent to the LRS

        :setter: Must be an HTTPResponse object
        :setter type: :class:`httplib.HTTPResponse`
        :rtype: :class:`httplib.HTTPResponse`
        """
        return self._response

    @response.setter
    def response(self, value):
        if value is not None and not isinstance(value, HTTPResponse):
            raise TypeError(
                f"Property 'response' in 'tincan.{self.__class__.__name__}' must be set with an HTTPResponse object"
            )
        self._response = value

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, value):
        """Setter for the _data attribute. Should be set from response.read()

        :param value: The body of the response object for the LRSResponse
        :type value: unicode
        """
        if value is not None and isinstance(value, (bytes, bytearray)):
            value = value.decode('utf-8')
        self._data = value

    @property
    def content(self):
        """Parsed content received from the LRS
        """
        return self._content

    @content.setter
    def content(self, value):
        self._content = value

    @content.deleter
    def content(self):
        del self._content



# Copyright 2014 Rustici Software
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from tincan.serializable_base import SerializableBase
from tincan.language_map import LanguageMap
from tincan.extensions import Extensions
from tincan.interaction_component_list import InteractionComponentList

"""
.. module:: activitydefinition
   :synopsis: The definition of the Activity object that provides \
   more granular details and extensions

"""


class ActivityDefinition(SerializableBase):
    _props = [
        'name',
        'description',
        'type',
        'more_info',
        'interaction_type',
        'correct_responses_pattern',
        'choices',
        'scale',
        'source',
        'target',
        'steps',
        'extensions',
    ]

    _interaction_types = [
        'choice',
        'sequencing',
        'likert',
        'matching',
        'performance',
        'true-false',
        'fill-in',
        'long-fill-in',
        'numeric',
        'other',
    ]

    def __init__(self, *args, **kwargs):
        self._name = None
        self._description = None
        self._type = None
        self._more_info = None
        self._interaction_type = None
        self._correct_responses_pattern = None
        self._choices = None
        self._scale = None
        self._source = None
        self._target = None
        self._steps = None
        self._extensions = None

        super(ActivityDefinition, self).__init__(*args, **kwargs)

    @property
    def name(self):
        """Name for Activity Definition

        :setter: Tries to convert to :class:`tincan.LanguageMap`
        :setter type: :class:`tincan.LanguageMap`
        :rtype: :class:`tincan.LanguageMap`

        """
        return self._name

    @name.setter
    def name(self, value):
        if value is not None and not isinstance(value, LanguageMap):
            value = LanguageMap(value)
        self._name = value

    @name.deleter
    def name(self):
        del self._name

    @property
    def description(self):
        """Description for Activity Definition

        :setter: Tries to convert to :class:`tincan.LanguageMap`
        :setter type: :class:`tincan.LanguageMap`
        :rtype: :class:`tincan.LanguageMap`

        """
        return self._description

    @description.setter
    def description(self, value):
        if value is not None and not isinstance(value, LanguageMap):
            value = LanguageMap(value)
        self._description = value

    @description.deleter
    def description(self):
        del self._description

    @property
    def type(self):
        """Type for Activity Definition

        :setter: Tries to convert to unicode
        :setter type: unicode
        :rtype: unicode

        """
        return self._type

    @type.setter
    def type(self, value):
        if value is not None:
            if value == '':
                raise ValueError("type cannot be set to an empty string")
            value = str(value)
        self._type = value

    @type.deleter
    def type(self):
        del self._type

    @property
    def more_info(self):
        """More info for Activity Definition

        :setter: Tries to convert to unicode
        :setter type: unicode
        :rtype: unicode

        """
        return self._more_info

    @more_info.setter
    def more_info(self, value):
        if value is not None:
            if value == '':
                raise ValueError("more_info cannot be set to an empty string")
            value = str(value)
        self._more_info = value

    @more_info.deleter
    def more_info(self):
        del self._more_info

    @property
    def interaction_type(self):
        """Interaction type for Activity Definition. Must be a value in interaction_types

        :setter: Tries to convert to unicode
        :setter type: unicode
        :rtype: unicode

        """
        return self._interaction_type

    @interaction_type.setter
    def interaction_type(self, value):
        if value is not None:
            if value not in self._interaction_types:
                raise ValueError("invalid interaction type")
        self._interaction_type = value

    @interaction_type.deleter
    def interaction_type(self):
        del self._interaction_type

    @property
    def correct_responses_pattern(self):
        """Correct responses pattern for Activity Definition

        :setter: Sets the correct responses pattern
        :setter type: list
        :rtype: list

        """
        return self._correct_responses_pattern

    @correct_responses_pattern.setter
    def correct_responses_pattern(self, value):
        if value is not None:
            if not isinstance(value, list):
                raise TypeError("correct responses pattern must be a list")
        self._correct_responses_pattern = value

    @correct_responses_pattern.deleter
    def correct_responses_pattern(self):
        del self._correct_responses_pattern

    @property
    def choices(self):
        """Choices for Activity Definition

        :setter: Tries to convert to :class:`tincan.InteractionComponentList`
        :setter type: :class:`tincan.InteractionComponentList`
        :rtype: :class:`tincan.InteractionComponentList`

        """
        return self._choices

    @choices.setter
    def choices(self, value):
        if value is not None and not isinstance(value, InteractionComponentList):
            value = InteractionComponentList(value)
        self._choices = value

    @choices.deleter
    def choices(self):
        del self._choices

    @property
    def scale(self):
        """Scale for Activity Definition

        :setter: Tries to convert to :class:`tincan.InteractionComponentList`
        :setter type: :class:`tincan.InteractionComponentList`
        :rtype: :class:`tincan.InteractionComponentList`

        """
        return self._scale

    @scale.setter
    def scale(self, value):
        if value is not None and not isinstance(value, InteractionComponentList):
            value = InteractionComponentList(value)
        self._scale = value

    @scale.deleter
    def scale(self):
        del self._scale

    @property
    def source(self):
        """Source for Activity Definition

        :setter: Tries to convert to :class:`tincan.InteractionComponentList`
        :setter type: :class:`tincan.InteractionComponentList`
        :rtype: :class:`tincan.InteractionComponentList`

        """
        return self._source

    @source.setter
    def source(self, value):
        if value is not None and not isinstance(value, InteractionComponentList):
            value = InteractionComponentList(value)
        self._source = value

    @source.deleter
    def source(self):
        del self._source

    @property
    def target(self):
        """Target for Activity Definition

        :setter: Tries to convert to :class:`tincan.InteractionComponentList`
        :setter type: :class:`tincan.InteractionComponentList`
        :rtype: :class:`tincan.InteractionComponentList`

        """
        return self._target

    @target.setter
    def target(self, value):
        if value is not None and not isinstance(value, InteractionComponentList):
            value = InteractionComponentList(value)
        self._target = value

    @target.deleter
    def target(self):
        del self._target

    @property
    def steps(self):
        """Steps for Activity Definition

        :setter: Tries to convert to :class:`tincan.InteractionComponentList`
        :setter type: :class:`tincan.InteractionComponentList`
        :rtype: :class:`tincan.InteractionComponentList`

        """
        return self._steps

    @steps.setter
    def steps(self, value):
        if value is not None and not isinstance(value, InteractionComponentList):
            value = InteractionComponentList(value)
        self._steps = value

    @steps.deleter
    def steps(self):
        del self._steps

    @property
    def extensions(self):
        """Extensions for Activity Definition

        :setter: Tries to convert to :class:`tincan.Extensions`
        :setter type: :class:`tincan.Extensions`
        :rtype: :class:`tincan.Extensions`

        """
        return self._extensions

    @extensions.setter
    def extensions(self, value):
        if value is not None and not isinstance(value, Extensions):
            value = Extensions(value)
        self._extensions = value

    @extensions.deleter
    def extensions(self):
        del self._extensions



# Copyright 2014 Rustici Software
#
# Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import uuid
import re
from datetime import datetime

from tincan.statement_base import StatementBase
from tincan.agent import Agent
from tincan.group import Group
from tincan.result import Result
from tincan.substatement import SubStatement
from tincan.statement_ref import StatementRef
from tincan.activity import Activity
from tincan.conversions.iso8601 import make_datetime
from tincan.version import Version


"""
.. module Statement
   :synopsis: A Statement object that contains all the information for one statement that is sent to an LRS
"""


class Statement(StatementBase):
    _UUID_REGEX = re.compile(
        r'^[a-f0-9]{8}-'
        r'[a-f0-9]{4}-'
        r'[1-5][a-f0-9]{3}-'
        r'[89ab][a-f0-9]{3}-'
        r'[a-f0-9]{12}$'
    )

    _props_req = [
        "id",
        "stored",
        "authority"
    ]

    _props = [
        "result",
        "version"
    ]

    _props.extend(StatementBase._props)
    _props.extend(_props_req)

    def __init__(self, *args, **kwargs):
        self._id = None
        self._stored = None
        self._authority = None
        self._result = None
        self._version = Version.latest

        super(Statement, self).__init__(*args, **kwargs)

    @property
    def id(self):
        """Id for Statement

        :setter: Tries to unicode
        :setter type: unicode
        :rtype: unicode

        """
        return self._id

    @id.setter
    def id(self, value):
        if value is not None and not isinstance(value, uuid.UUID):
            if isinstance(value, str) and not self._UUID_REGEX.match(value):
                raise ValueError("Invalid UUID string")
            value = uuid.UUID(value)
        self._id = value

    @id.deleter
    def id(self):
        del self._id

    @property
    def object(self):
        """Object for Statement

        :setter: Sets the object
        :setter type: :class:`tincan.Agent` | :class:`tincan.Group` | :class:`tincan.StatementRef` |
                      :class:`tincan.SubStatement` | :class:`tincan.Activity`
        :rtype: :class:`tincan.Agent` | :class:`tincan.Group` | :class:`tincan.StatementRef` |
                :class:`tincan.SubStatement` | :class:`tincan.Activity`

        """
        return self._object

    @object.setter
    def object(self, value):
        if value is not None:
            if not isinstance(value, Agent) and \
                    not isinstance(value, Group) and \
                    not isinstance(value, SubStatement) and \
                    not isinstance(value, StatementRef) and \
                    not isinstance(value, Activity):
                if isinstance(value, dict):
                    if 'object_type' in value or 'objectType' in value:
                        if 'objectType' in value:
                            value['object_type'] = value['objectType']
                            value.pop('objectType')
                        if value['object_type'] == 'Agent':
                            value = Agent(value)
                        elif value['object_type'] == 'SubStatement':
                            value = SubStatement(value)
                        elif value['object_type'] == 'StatementRef':
                            value = StatementRef(value)
                        elif value['object_type'] == 'Activity':
                            value = Activity(value)
                        elif value['object_type'] == 'Group':
                            value = Group(value)
                        else:
                            value = Activity(value)
                    else:
                        value = Activity(value)
        self._object = value

    @object.deleter
    def object(self):
        del self._object

    @property
    def stored(self):
        """Storage time. Tries to convert to :class:`datetime.datetime`.
        If no timezone is given, makes a naive `datetime.datetime`.

        Strings will be parsed as ISO 8601 timestamps.

        If a number is provided, it will be interpreted as a UNIX
        timestamp, which by definition is UTC.

        If a `dict` is provided, does `datetime.datetime(**value)`.

        If a `tuple` or a `list` is provided, does
        `datetime.datetime(*value)`. Uses the timezone in the tuple or
        list if provided.

        :setter type: :class:`datetime.datetime` | unicode | str | int | float | dict | tuple | list | None
        :rtype: :class:`datetime.datetime`
        """
        return self._stored

    @stored.setter
    def stored(self, value):
        if value is None or isinstance(value, datetime):
            self._stored = value
            return

        try:
            self._stored = make_datetime(value)
        except TypeError as e:
            message = (
                f"Property 'stored' in a 'tincan.{self.__class__.__name__}' "
                f"object must be set with a "
                f"datetime.datetime, str, unicode, int, float, dict "
                f"or None.\n\n{repr(e)}"
            )
            raise TypeError(message) from e


    @stored.deleter
    def stored(self):
        del self._stored

    @property
    def authority(self):
        """Authority for Statement

        :setter: Tries to convert to :class:`tincan.Agent`
        :setter type: :class:`tincan.Agent`
        :rtype: :class:`tincan.Agent`

        """
        return self._authority

    @authority.setter
    def authority(self, value):
        if value is not None and not isinstance(value, Agent):
            value = Agent(value)
        self._authority = value

    @authority.deleter
    def authority(self):
        del self._authority

    @property
    def result(self):
        """Result for Statement

        :setter: Tries to convert to :class:`tincan.Result`
        :setter type: :class:`tincan.Result`
        :rtype: :class:`tincan.Result`

        """
        return self._result

    @result.setter
    def result(self, value):
        if value is not None and not isinstance(value, Result):
            value = Result(value)
        self._result = value

    @result.deleter
    def result(self):
        del self._result

    @property
    def version(self):
        """Version for Statement

        :setter: Tries to convert to unicode
        :setter type: unicode
        :rtype: unicode

        """
        return self._version

    @version.setter
    def version(self, value):
        if value is not None:
            if value == '':
                raise ValueError("Property version can not be set to an empty string")
            elif not isinstance(value, str):
                value = str(value)
        self._version = value

    @version.deleter
    def version(self):
        del self._version



# Copyright 2014 Rustici Software
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import unittest

if __name__ == '__main__':
    from test.main import setup_tincan_path

    setup_tincan_path()
from tincan import Version


class TinCanBaseTestCase(unittest.TestCase):
    # PEP8 says this should be lowercase, but unittest breaks this rule
    def assertSerializeDeserialize(self, obj, version=None):
        """
        Verifies that objects are packed to JSON and unpacked
        correctly for the version specified. If no version is
        specified, tests for all versions in TCAPIVersion.
        :param obj: A TinCan object to be tested.
        :param version: The version according to whose schema we will test.
        :type version: str
        """
        tested_versions = [version] if version is not None else Version.supported
        for version in tested_versions:
            constructor = obj.__class__.from_json
            json_obj = obj.to_json(version)
            clone = constructor(json_obj)

            self.assertEqual(obj.__class__, clone.__class__)

            if isinstance(obj, dict):
                orig_dict = obj
                clone_dict = clone
            else:
                orig_dict = obj.__dict__
                clone_dict = clone.__dict__

            self.assertEqual(orig_dict, clone_dict)



"""
Client library for communicating with an LRS (Learning Record Store)
implementing Tin Can API version 1.0.0 or 1.0.1.

Web site: <https://github.com/RusticiSoftware/TinCanPython>

For more info about the Tin Can API, see <http://tincanapi.com/>.
"""

# These imports are for convenience to external modules only.
# Internal tincan modules should continue to use explicit imports,
# since this file will not have run yet.
#
# For example, from the outside, you can say:
# from tincan import RemoteLRS, LRSResponse
#
# but inside the tincan package, we have to use:
#    from tincan.remote_lrs import RemoteLRS
#    from tincan.lrs_response import LRSResponse

from tincan.about import About
from tincan.activity import Activity
from tincan.activity_definition import ActivityDefinition
from tincan.activity_list import ActivityList
from tincan.agent import Agent
from tincan.agent_account import AgentAccount
from tincan.agent_list import AgentList
from tincan.attachment import Attachment
from tincan.attachment_list import AttachmentList
from tincan.base import Base
from tincan.context import Context
from tincan.context_activities import ContextActivities
from tincan.documents.activity_profile_document import ActivityProfileDocument
from tincan.documents.agent_profile_document import AgentProfileDocument
from tincan.documents.document import Document
from tincan.documents.state_document import StateDocument
from tincan.extensions import Extensions
from tincan.group import Group
from tincan.http_request import HTTPRequest
from tincan.interaction_component import InteractionComponent
from tincan.interaction_component_list import InteractionComponentList
from tincan.language_map import LanguageMap
from tincan.lrs_response import LRSResponse
from tincan.remote_lrs import RemoteLRS
from tincan.result import Result
from tincan.score import Score
from tincan.serializable_base import SerializableBase
from tincan.statement import Statement
from tincan.statement_base import StatementBase
from tincan.statement_list import StatementList
from tincan.statement_ref import StatementRef
from tincan.statement_targetable import StatementTargetable
from tincan.statements_result import StatementsResult
from tincan.substatement import SubStatement
from tincan.typed_list import TypedList
from tincan.verb import Verb
from tincan.version import Version



# Copyright 2014 Rustici Software
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
Functions for converting ``datetime.timedelta`` and
``datetime.datetime`` to and from ISO 8601 strings.
"""

import datetime
# struct_time does not preserve millisecond accuracy per
# Tin Can spec, so this is disabled to discourage its use.
# from time import mktime, struct_time
import aniso8601
from pytz import utc

def make_timedelta(value):
    """Tries to convert the given value to a :class:`datetime.timedelta`.

    Strings will be parsed as ISO 8601 durations.

    If a number is provided, it will be interpreted as the number of
    seconds.

    If a `dict` is provided, does `datetime.timedelta(**value)`.

    :param value: something to convert
    :type value: str | unicode | float | int | datetime.timedelta | dict
    :return: the value after conversion
    :rtype: datetime.timedelta

    """

    if isinstance(value, str):
        try:
            return aniso8601.parse_duration(value)
        except Exception as e:
            msg = (
                f"Conversion to datetime.timedelta failed. Could not "
                f"parse the given string as an ISO 8601 duration: "
                f"{repr(value)}\n\n"
                f"{repr(e)}"
            )
            raise ValueError(msg)

    try:
        if isinstance(value, datetime.timedelta):
            return value
        elif isinstance(value, dict):
            return datetime.timedelta(**value)
        elif isinstance(value, (float, int)):
            return datetime.timedelta(seconds=value)
        else:
            return datetime.timedelta(value)
    except Exception as e:
        msg = (
            f"Could not convert the given value of type '{value.__class__.__name__}' to a "
            f"datetime.timedelta: {repr(value)}\n\n"
            f"{repr(e)}"
        )
        raise TypeError(msg) if isinstance(e, TypeError) else ValueError(msg)


def jsonify_timedelta(value):
    """Converts a `datetime.timedelta` to an ISO 8601 duration
    string for JSON-ification.

    :param value: something to convert
    :type value: datetime.timedelta
    :return: the value after conversion
    :rtype unicode

    """

    assert isinstance(value, datetime.timedelta)

    # split seconds to larger units
    seconds = value.total_seconds()
    minutes, seconds = divmod(seconds, 60)
    hours, minutes = divmod(minutes, 60)
    days, hours = divmod(hours, 24)

    days, hours, minutes = list(map(int, (days, hours, minutes)))
    seconds = round(seconds, 6)

    # build date
    date = ''
    if days:
        date = '%sD' % days

    # build time
    time = 'T'

    # hours
    bigger_exists = date or hours
    if bigger_exists:
        time += '{:02}H'.format(hours)

    # minutes
    bigger_exists = bigger_exists or minutes
    if bigger_exists:
        time += '{:02}M'.format(minutes)

    # seconds
    if seconds.is_integer():
        seconds = '{:02}'.format(int(seconds))
    else:
        # 9 chars long w/leading 0, 6 digits after decimal
        seconds = '%09.6f' % seconds
        # remove trailing zeros
        seconds = seconds.rstrip('0')

    time += '{}S'.format(seconds)

    return 'P' + date + time


def make_datetime(value):
    """Tries to convert the given value to a :class:`datetime.datetime`. If
    no timezone is given, raises a ValueError.

    Strings will be parsed as ISO 8601 timestamps.

    If a number is provided, it will be interpreted as a UNIX
    timestamp, which by definition is UTC.

    If a `dict` is provided, does `datetime.datetime(**value)`.

    If a `tuple` or a `list` is provided, does
    `datetime.datetime(*value)`. Uses the timezone in the tuple or
    list if provided.

    :param value: something to convert
    :type value: str | unicode | float | int | :class:`datetime.datetime` | dict | list | tuple
    :return: the value after conversion
    :rtype: :class:`datetime.datetime`
    :raises: ValueError | TypeError

    """
    result = _make_datetime(value)
    if not result.tzinfo:
        raise ValueError(
            f"value was a timestamp, but no timezone was set! "
            f"Value was a '{value.__class__.__name__}' object: {repr(value)}"
            f"\n\n"
            f"Converted to naive 'datetime.datetime' object: {repr(result)}"
        )

    return result


def _make_datetime(value):
    """Helper function for `make_datetime()`.

    Tries to convert the given value to a
    :class:`datetime.datetime`. But, unlike make_datetime(), if no
    timezone is given, makes a naive `datetime.datetime`.

    Strings will be parsed as ISO 8601 timestamps.

    If a number is provided, it will be interpreted as a UNIX
    timestamp, which by definition is UTC.

    If a `dict` is provided, does `datetime.datetime(**value)`.

    If a `tuple` or a `list` is provided, does
    `datetime.datetime(*value)`. Uses the timezone in the tuple or
    list if provided.

    :param value: something to convert
    :type value: str | unicode | float | int | :class:`datetime.datetime` | dict | list | tuple
    :return: the value after conversion
    :rtype: :class:`datetime.datetime`
    :raises: ValueError | TypeError
    """

    if isinstance(value, str):
        try:
            return aniso8601.parse_datetime(value)
        except Exception as e:
            raise ValueError(
                f"Conversion to datetime.datetime failed. Could not "
                f"parse the given string as an ISO 8601 timestamp: "
                f"{repr(value)}\n\n"
                f"{repr(e)}"
            )

    try:
        if isinstance(value, datetime.datetime):
            return value
        elif isinstance(value, dict):
            tzinfo = value.pop('tzinfo', None)
            if tzinfo:
                return tzinfo.localize(datetime.datetime(**value))
            else:
                return datetime.datetime(**value)
        # struct_time does not preserve millisecond accuracy per
        # TinCan spec, so this is disabled to discourage its use.
        # elif isinstance(value, struct_time):
        #     posix = mktime(value)
        #     return datetime.datetime.utcfromtimestamp(posix).replace(tzinfo=utc)
        elif isinstance(value, (tuple, list)):
            return tuple_to_datetime(value)
        else:
            return datetime.datetime.utcfromtimestamp(value).replace(tzinfo=utc)
    except Exception as e:
        msg = (
            f"Could not convert the given value of type '{value.__class__.__name__}' to a "
            f"datetime.datetime: {repr(value)}\n\n"
            f"{repr(e)}"
        )
        raise TypeError(msg) if isinstance(e, TypeError) else ValueError(msg)


def jsonify_datetime(value):
    assert isinstance(value, datetime.datetime)
    return value.isoformat()


def tuple_to_datetime(value):
    # look at the last value and see if it might be a tzinfo
    tzinfo = value[-1]
    tzinfo = tzinfo if hasattr(tzinfo, 'localize') else None

    if tzinfo:
        try:
            args = value[:-1]
            return tzinfo.localize(datetime.datetime(*args))
        except Exception as e:
            raise ValueError(
                f"Failed to call tzinfo.localize(datetime) method "
                f"of tzinfo object: {repr(tzinfo)}\n"
                f"tuple to convert to datetime.datetime was: {repr(value)}"
                f"\n\n"
                f"{repr(e)}"
            )
    else:
        return datetime.datetime(*value)



# Copyright 2014 Rustici Software
#
# Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from datetime import timedelta

from tincan.serializable_base import SerializableBase
from tincan.score import Score
from tincan.extensions import Extensions
from tincan.conversions.iso8601 import make_timedelta


class Result(SerializableBase):
    """Stores the state of an activity.

    Can be created from a dict, another Result, or from kwargs.

    All these attributes are optional and settable to None:

    :param score: Contains the score and its scaling information
    :type score: :class:`tincan.Score`
    :param success: Whether successful
    :type success: bool
    :param completion: Whether completed
    :type completion: bool
    :param duration: How long it took
    :type duration: timedelta
    :param response: HTTPResponse data
    :type response: unicode
    :param extensions: Custom user data
    :type extensions: :class:`tincan.Extensions`
    """

    _props = [
        'score',
        'success',
        'completion',
        'duration',
        'response',
        'extensions',
    ]

    def __init__(self, *args, **kwargs):
        self._score = None
        self._success = None
        self._completion = None
        self._duration = None
        self._response = None
        self._extensions = None

        super(Result, self).__init__(*args, **kwargs)

    @property
    def score(self):
        """Score for Result

        :setter: Tries to convert to :class:`tincan.Score`. If
        None is provided, this signifies the absence of this data.
        :setter type: :class:`tincan.Score` | dict | None
        :rtype: :class:`tincan.Score` | None

        """
        return self._score

    @score.setter
    def score(self, value):
        try:
            self._score = value if value is None or isinstance(value, Score) else Score(value)
        except Exception as e:
            msg = (
                f"Property 'score' in 'tincan.{self.__class__.__name__}' object: could not create a"
                f" 'tincan.Score' object from value: {repr(value)}\n\n"
            )
            msg += repr(e)
            e_type = TypeError if not isinstance(value, dict) else ValueError
            raise e_type(msg) from e

    @score.deleter
    def score(self):
        del self._score

    @property
    def success(self):
        """Success for Result

        :setter: Tries to convert to bool. If None is provided,
        this signifies the absence of this data.
        :setter type: bool | None
        :rtype: bool | None

        """
        return self._success

    @success.setter
    def success(self, value):
        self._success = value if value is None else bool(value)

    @success.deleter
    def success(self):
        del self._success

    @property
    def completion(self):
        """Completion for Result

        :setter: Tries to convert to bool. If None is provided,
        this signifies the absence of this data.
        :setter type: bool | None
        :rtype: bool | None

        """
        return self._completion

    @completion.setter
    def completion(self, value):
        self._completion = value if value is None else bool(value)

    @completion.deleter
    def completion(self):
        del self._completion

    @property
    def duration(self):
        """Duration for Result

        :setter: Tries to convert to :class:`datetime.timedelta`. If
        None is provided, this signifies the absence of this data.

        Strings will be parsed as ISO 8601 durations.

        If a number is provided, it will be interpreted as the number of
        seconds.

        If a `dict` is provided, does `datetime.timedelta(**value)`.

        :setter type: :class:`datetime.timedelta` | unicode | str | int | float | dict | None
        :rtype: :class:`datetime.timedelta` | None
        :raises: ValueError if the provided data was a valid type, but could not be converted
        :raises: TypeError if unsupported type is provided
        """
        return self._duration

    @duration.setter
    def duration(self, value):
        if value is None or isinstance(value, timedelta):
            self._duration = value
            return

        try:
            self._duration = make_timedelta(value)
        except Exception as e:
            message = (
                f"Property 'duration' in a 'tincan.{self.__class__.__name__}' object must be set with a "
                f"datetime.timedelta, str, unicode, int, float or None.\n\n{repr(e)}"
            )
            raise Exception(message) from e

    @duration.deleter
    def duration(self):
        del self._duration

    @property
    def response(self):
        """Response for Result

        :setter: Tries to convert to unicode. If None is provided,
        this signifies the absence of this data.
        :setter type: unicode | str | None
        :rtype: unicode | None

        """
        return self._response

    @response.setter
    def response(self, value):
        try:
            self._response = value if value is None else str(value)
        except Exception as e:
            e_type = ValueError if isinstance(value, (list, tuple)) else TypeError
            msg = (
                f"Property 'response' in a 'tincan.{self.__class__.__name__}' object must be set with a "
                f"string, unicode or None.\n\n"
            )
            msg += repr(e)
            raise e_type(msg) from e

    @response.deleter
    def response(self):
        del self._response

    @property
    def extensions(self):
        """Extensions for Result

        :setter: Tries to convert to :class:`tincan.Extensions`. If None is provided,
        this signifies the absence of this data.
        :setter type: :class:`tincan.Extensions` | dict | None
        :rtype: :class:`tincan.Extensions` | None

        """
        return self._extensions

    @extensions.setter
    def extensions(self, value):
        if value is None or isinstance(value, Extensions):
            self._extensions = value
            return
        try:
            self._extensions = Extensions(value)
        except Exception as e:
            e_type = ValueError if isinstance(value, dict) else TypeError
            msg = (
                f"Property 'extensions' in 'tincan.{self.__class__.__name__}' object: could not create a"
                f" 'tincan.Extensions' object from value: {repr(value)}\n\n"
            )
            msg += repr(e)
            raise e_type(msg)

    @extensions.deleter
    def extensions(self):
        del self._extensions



# Copyright 2014 Rustici Software
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from tincan.serializable_base import SerializableBase
from tincan.statement_targetable import StatementTargetable
from tincan.activity_definition import ActivityDefinition

"""
.. module:: activity
   :synopsis: The Activity object that defines the boundaries on the 'Object' \
   part of 'Actor-Verb-Object' structure of a Statement

"""


class Activity(SerializableBase, StatementTargetable):
    _props_req = [
        'id',
        'object_type'
    ]

    _props = [
        'definition',
    ]

    _props.extend(_props_req)

    def __init__(self, *args, **kwargs):
        self._id = None
        self._object_type = None
        self._definition = None

        super(Activity, self).__init__(*args, **kwargs)

    @property
    def id(self):
        """Id for Activity

        :setter: Sets the id
        :setter type: unicode
        :rtype: unicode

        """
        return self._id

    @id.setter
    def id(self, value):
        if value is not None:
            if value == '':
                raise ValueError(
                    f"Property 'id' in 'tincan.{self.__class__.__name__}' object must be not empty."
                )
        self._id = None if value is None else str(value)

    @property
    def object_type(self):
        """Object type for Activity. Will always be "Activity"

        :setter: Tries to convert to unicode
        :setter type: unicode
        :rtype: unicode

        """
        return self._object_type

    @object_type.setter
    def object_type(self, _):
        self._object_type = 'Activity'

    @property
    def definition(self):
        """Definition for Activity

        :setter: Tries to convert to :class:`tincan.ActivityDefinition`
        :setter type: :class:`tincan.ActivityDefinition`
        :rtype: :class:`tincan.ActivityDefinition`

        """
        return self._definition

    @definition.setter
    def definition(self, value):
        if value is not None and not isinstance(value, ActivityDefinition):
            value = ActivityDefinition(value)
        self._definition = value

    @definition.deleter
    def definition(self):
        del self._definition



# Copyright 2014 Rustici Software
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from tincan.serializable_base import SerializableBase
from tincan.activity_list import ActivityList
from tincan.activity import Activity


class ContextActivities(SerializableBase):
    _props = [
        'category',
        'parent',
        'grouping',
        'other',
    ]

    def __init__(self, *args, **kwargs):
        self._category = None
        self._parent = None
        self._grouping = None
        self._other = None

        super(ContextActivities, self).__init__(*args, **kwargs)

    @property
    def category(self):
        """Category for Context Activities

        :setter: Tries to convert to :class:`tincan.ActivityList`
        :setter type: :class:`tincan.ActivityList`
        :rtype: :class:`tincan.ActivityList`

        """
        return self._category

    @category.setter
    def category(self, value):
        value = self._activity_or_list(value)
        self._category = value

    @category.deleter
    def category(self):
        del self._category

    @property
    def parent(self):
        """Parent for Context Activities

        :setter: Tries to convert to :class:`tincan.ActivityList`
        :setter type: :class:`tincan.ActivityList`
        :rtype: :class:`tincan.ActivityList`

        """
        return self._parent

    @parent.setter
    def parent(self, value):
        value = self._activity_or_list(value)
        self._parent = value

    @parent.deleter
    def parent(self):
        del self._parent

    @property
    def grouping(self):
        """Grouping for Context Activities

        :setter: Tries to convert to :class:`tincan.ActivityList`
        :setter type: :class:`tincan.ActivityList`
        :rtype: :class:`tincan.ActivityList`

        """
        return self._grouping

    @grouping.setter
    def grouping(self, value):
        value = self._activity_or_list(value)
        self._grouping = value

    @grouping.deleter
    def grouping(self):
        del self._grouping

    @property
    def other(self):
        """Other for Context Activities

        :setter: Tries to convert to :class:`tincan.ActivityList`
        :setter type: :class:`tincan.ActivityList`
        :rtype: :class:`tincan.ActivityList`

        """
        return self._other

    @other.setter
    def other(self, value):
        value = self._activity_or_list(value)
        self._other = value

    @other.deleter
    def other(self):
        del self._other

    @staticmethod
    def _activity_or_list(value):
        """Tries to convert value to :class:`tincan.ActivityList`

        :setter type: :class:`tincan.ActivityList`
        :rtype: :class:`tincan.ActivityList`
        """
        result = value
        if value is not None and not isinstance(value, ActivityList):
            try:
                result = ActivityList([Activity(value)])
            except (TypeError, AttributeError):
                result = ActivityList(value)
        return result



# Copyright 2014 Rustici Software
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from tincan.serializable_base import SerializableBase

"""
.. module:: languagemap
   :synopsis: A simple wrapper for a map containing language mappings

"""


class LanguageMap(dict, SerializableBase):
    def __init__(self, *args, **kwargs):
        """Initializes a LanguageMap with the given mapping

        This constructor will first check the arguments for flatness
        to avoid nested languagemaps (which are invalid) and then
        call the base dict constructor

        """
        check_args = dict(*args, **kwargs)
        list(map(lambda k_v: (k_v[0], self._check_basestring(k_v[1])), iter(check_args.items())))
        super(LanguageMap, self).__init__(check_args)

    def __setitem__(self, prop, value):
        """Allows bracket notation for setting values with hyphenated keys

        :param prop: The property to set
        :type prop: str
        :param value: The value to set
        :type value: obj

        :raises: NameError, LanguageMapTypeError

        """
        self._check_basestring(value)
        super(LanguageMap, self).__setitem__(prop, value)

    @staticmethod
    def _check_basestring(value):
        """Ensures that value is an instance of basestring

        :param value: the value to check
        :type value: any

        """
        if not isinstance(value, str):
            raise TypeError("Value must be a stringstring_types")



# Copyright 2014 Rustici Software
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
.. module:: statement_targetable
   :synopsis: Provides a way to define objects as targetable by a statement

   At this time, objects that are targetable need access to only the \
   object type, but this may change in the future

"""


class StatementTargetable(object):
    def __init__(self):
        self.object_type = None

    def get_object_type(self):
        """Returns the object type of self [Activity | Agent | \
        StatementRef | SubStatement]
        """
        return self.object_type



# Copyright 2014 Rustici Software
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from tincan.documents import Document
from tincan.agent import Agent


class AgentProfileDocument(Document):
    """Extends :class:`tincan.Document` with an Agent field, can be created from a dict,
    another :class:`tincan.Document`, or from kwargs.

    :param id: The id of this document
    :type id: unicode
    :param content_type: The content_type of the content of this document
    :type content_type: unicode
    :param content: The content of this document
    :type content: bytearray
    :param etag: The etag of this document
    :type etag: unicode
    :param timestamp: The timestamp of this document
    :type timestamp: :class:`datetime.datetime`
    :param agent: The agent object of this document
    :type agent: :class:`tincan.Agent`
    """

    _props_req = list(Document._props_req)

    _props_req.extend([
        'agent',
    ])

    _props = list(Document._props)

    _props.extend(_props_req)

    def __init__(self, *args, **kwargs):
        self._agent = None
        super(AgentProfileDocument, self).__init__(*args, **kwargs)

    @property
    def agent(self):
        """The Document's agent object

        :setter: Tries to convert to :class:`tincan.Agent`
        :setter type: :class:`tincan.Agent`
        :rtype: :class:`tincan.Agent`
        """
        return self._agent

    @agent.setter
    def agent(self, value):
        if not isinstance(value, Agent) and value is not None:
            try:
                value = Agent(value)
            except:
                raise TypeError(
                    f"Property 'agent' in 'tincan.{self.__class__.__name__}' must be set with a type "
                    f"that can be constructed into an tincan.Agent object."
                )
        self._agent = value



# Copyright 2014 Rustici Software
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import http.client
import json
import base64


from urllib.parse import urlparse, urlencode

from tincan.lrs_response import LRSResponse
from tincan.http_request import HTTPRequest
from tincan.statement_list import StatementList
from tincan.agent import Agent
from tincan.statement import Statement
from tincan.activity import Activity
from tincan.statements_result import StatementsResult
from tincan.about import About
from tincan.version import Version
from tincan.base import Base
from tincan.documents import (
    StateDocument,
    ActivityProfileDocument,
    AgentProfileDocument
)

"""
.. module:: remote_lrs
   :synopsis: The RemoteLRS class implements LRS communication.
"""


class RemoteLRS(Base):
    _props_req = [
        'version',
        'endpoint',
        'auth',
    ]

    _props = []

    _props.extend(_props_req)

    def __init__(self, *args, **kwargs):
        """RemoteLRS Constructor

        :param endpoint: lrs endpoint
        :type endpoint: str | unicode
        :param version: Version used for lrs communication
        :type version: str | unicode
        :param username: Username for lrs. Used to build the authentication string.
        :type username: str | unicode
        :param password: Password for lrs. Used to build the authentication string.
        :type password: str | unicode
        :param auth: Authentication string
        :type auth: str | unicode
        """

        self._version = Version.latest
        self._endpoint = None
        self._auth = None

        if "username" in kwargs \
                and kwargs["username"] is not None \
                and "password" in kwargs \
                and kwargs["password"] is not None \
                and "auth" not in kwargs:
            bytes = (
                str(kwargs["username"]) +
                ":" +
                str(kwargs["password"])).encode('ASCII')
            auth_string = str.format("Basic {}", base64.b64encode(bytes).decode("utf-8"))

            kwargs.pop("username")
            kwargs.pop("password")
            kwargs["auth"] = auth_string

        super(RemoteLRS, self).__init__(*args, **kwargs)

    def _send_request(self, request):
        """Establishes connection and returns http response based off of request.

        :param request: HTTPRequest object
        :type request: :class:`tincan.http_request.HTTPRequest`
        :returns: LRS Response object
        :rtype: :class:`tincan.lrs_response.LRSResponse`
        """
        headers = {"X-Experience-API-Version": self.version}

        if self.auth is not None:
            headers["Authorization"] = self.auth

        headers.update(request.headers)

        params = request.query_params
        params = {k: str(params[k]).encode('utf-8') for k in list(params.keys())}
        params = urlencode(params)

        if request.resource.startswith('http'):
            url = request.resource
        else:
            url = self.endpoint
            url += request.resource

        parsed = urlparse(url)

        if parsed.scheme == "https":
            web_req = http.client.HTTPSConnection(parsed.hostname, parsed.port)
        else:
            web_req = http.client.HTTPConnection(parsed.hostname, parsed.port)

        path = parsed.path
        if parsed.query or parsed.path:
            path += "?"
            if parsed.query:
                path += parsed.query
            if params:
                path += params

        if hasattr(request, "content") and request.content is not None:
            web_req.request(
                method=request.method,
                url=path,
                body=request.content,
                headers=headers,
            )
        else:
            web_req.request(
                method=request.method,
                url=path,
                headers=headers,
            )

        response = web_req.getresponse()
        data = response.read()
        web_req.close()

        if (200 <= response.status < 300
            or (response.status == 404
                and hasattr(request, "ignore404")
                and request.ignore404)):
            success = True
        else:
            success = False

        return LRSResponse(
            success=success,
            request=request,
            response=response,
            data=data,
        )

    def about(self):
        """Gets about response from LRS

        :return: LRS Response object with the returned LRS about object as content
        :rtype: :class:`tincan.lrs_response.LRSResponse`
        """
        request = HTTPRequest(
            method="GET",
            resource="about"
        )
        lrs_response = self._send_request(request)

        if lrs_response.success:
            lrs_response.content = About.from_json(lrs_response.data)

        return lrs_response

    def save_statement(self, statement):
        """Save statement to LRS and update statement id if necessary

        :param statement: Statement object to be saved
        :type statement: :class:`tincan.statement.Statement`
        :return: LRS Response object with the saved statement as content
        :rtype: :class:`tincan.lrs_response.LRSResponse`
        """
        if not isinstance(statement, Statement):
            statement = Statement(statement)

        request = HTTPRequest(
            method="POST",
            resource="statements"
        )
        if statement.id is not None:
            request.method = "PUT"
            request.query_params["statementId"] = statement.id

        request.headers["Content-Type"] = "application/json"
        request.content = statement.to_json(self.version)

        lrs_response = self._send_request(request)

        if lrs_response.success:
            if statement.id is None:
                statement.id = json.loads(lrs_response.data)[0]
            lrs_response.content = statement

        return lrs_response

    def save_statements(self, statements):
        """Save statements to LRS and update their statement id's

        :param statements: A list of statement objects to be saved
        :type statements: :class:`StatementList`
        :return: LRS Response object with the saved list of statements as content
        :rtype: :class:`tincan.lrs_response.LRSResponse`
        """
        if not isinstance(statements, StatementList):
            statements = StatementList(statements)

        request = HTTPRequest(
            method="POST",
            resource="statements"
        )
        request.headers["Content-Type"] = "application/json"

        request.content = statements.to_json()

        lrs_response = self._send_request(request)

        if lrs_response.success:
            id_list = json.loads(lrs_response.data)
            for s, statement_id in zip(statements, id_list):
                s.id = statement_id

            lrs_response.content = statements

        return lrs_response

    def retrieve_statement(self, statement_id):
        """Retrieve a statement from the server from its id

        :param statement_id: The UUID of the desired statement
        :type statement_id: str | unicode
        :return: LRS Response object with the retrieved statement as content
        :rtype: :class:`tincan.lrs_response.LRSResponse`
        """
        request = HTTPRequest(
            method="GET",
            resource="statements"
        )
        request.query_params["statementId"] = statement_id

        lrs_response = self._send_request(request)

        if lrs_response.success:
            lrs_response.content = Statement.from_json(lrs_response.data)

        return lrs_response

    def retrieve_voided_statement(self, statement_id):
        """Retrieve a voided statement from the server from its id

        :param statement_id: The UUID of the desired voided statement
        :type statement_id: str | unicode
        :return: LRS Response object with the retrieved voided statement as content
        :rtype: :class:`tincan.lrs_response.LRSResponse`
        """
        request = HTTPRequest(
            method="GET",
            resource="statements"
        )
        request.query_params["voidedStatementId"] = statement_id

        lrs_response = self._send_request(request)

        if lrs_response.success:
            lrs_response.content = Statement.from_json(lrs_response.data)

        return lrs_response

    def query_statements(self, query):
        """Query the LRS for statements with specified parameters

        :param query: Dictionary of query parameters and their values
        :type query: dict
        :return: LRS Response object with the returned StatementsResult object as content
        :rtype: :class:`tincan.lrs_response.LRSResponse`

        .. note::
           Optional query parameters are\n
               **statementId:** (*str*) ID of the Statement to fetch
               **voidedStatementId:** (*str*) ID of the voided Statement to fetch
               **agent:** (*Agent* |*Group*) Filter to return Statements for which the
               specified Agent or Group is the Actor
               **verb:** (*Verb id IRI*) Filter to return Statements matching the verb id
               **activity:** (*Activity id IRI*) Filter to return Statements for which the
               specified Activity is the Object
               **registration:** (*UUID*) Filter to return Statements matching the specified registration ID
               **related_activities:** (*bool*) Include Statements for which the Object,
               Context Activities or any Sub-Statement
               properties match the specified Activity
               **related_agents:** (*bool*) Include Statements for which the Actor, Object,
               Authority, Instructor, Team, or any Sub-Statement properties match the specified Agent
               **since:** (*datetime*) Filter to return Statements stored since the specified datetime
               **until:** (*datetime*) Filter to return Statements stored at or before the specified datetime
               **limit:** (*positive int*) Allow <limit> Statements to be returned. 0 indicates the
               maximum supported by the LRS
               **format:** (*str* {"ids"|"exact"|"canonical"}) Manipulates how the LRS handles
               importing and returning the statements
               **attachments:** (*bool*) If true, the LRS will use multipart responses and include
               all attachment data per Statement returned.
               Otherwise, application/json is used and no attachment information will be returned
               **ascending:** (*bool*) If true, the LRS will return results in ascending order of
               stored time (oldest first)
        """
        params = {}

        param_keys = [
            "registration",
            "since",
            "until",
            "limit",
            "ascending",
            "related_activities",
            "related_agents",
            "format",
            "attachments",
        ]

        for k, v in query.items():
            if v is not None:
                if k == "agent":
                    params[k] = v.to_json(self.version)
                elif k == "verb" or k == "activity":
                    params[k] = v.id
                elif k in param_keys:
                    params[k] = v

        request = HTTPRequest(
            method="GET",
            resource="statements"
        )
        request.query_params = params

        lrs_response = self._send_request(request)

        if lrs_response.success:
            lrs_response.content = StatementsResult.from_json(lrs_response.data)

        return lrs_response

    def more_statements(self, more_url):
        """Query the LRS for more statements

        :param more_url: URL from a StatementsResult object used to retrieve more statements
        :type more_url: str | unicode
        :return: LRS Response object with the returned StatementsResult object as content
        :rtype: :class:`tincan.lrs_response.LRSResponse`
        """
        if isinstance(more_url, StatementsResult):
            more_url = more_url.more

        more_url = self.get_endpoint_server_root() + more_url

        request = HTTPRequest(
            method="GET",
            resource=more_url
        )

        lrs_response = self._send_request(request)

        if lrs_response.success:
            lrs_response.content = StatementsResult.from_json(lrs_response.data)

        return lrs_response

    def retrieve_state_ids(self, activity, agent, registration=None, since=None):
        """Retrieve state id's from the LRS with the provided parameters

        :param activity: Activity object of desired states
        :type activity: :class:`tincan.activity.Activity`
        :param agent: Agent object of desired states
        :type agent: :class:`tincan.agent.Agent`
        :param registration: Registration UUID of desired states
        :type registration: str | unicode
        :param since: Retrieve state id's since this time
        :type since: str | unicode
        :return: LRS Response object with the retrieved state id's as content
        :rtype: :class:`tincan.lrs_response.LRSResponse`
        """
        if not isinstance(activity, Activity):
            activity = Activity(activity)

        if not isinstance(agent, Agent):
            agent = Agent(agent)

        request = HTTPRequest(
            method="GET",
            resource="activities/state"
        )
        request.query_params = {
            "activityId": activity.id,
            "agent": agent.to_json(self.version)
        }

        if registration is not None:
            request.query_params["registration"] = registration
        if since is not None:
            request.query_params["since"] = since

        lrs_response = self._send_request(request)

        if lrs_response.success:
            lrs_response.content = json.loads(lrs_response.data)

        return lrs_response

    def retrieve_state(self, activity, agent, state_id, registration=None):
        """Retrieve state from LRS with the provided parameters

        :param activity: Activity object of desired state
        :type activity: :class:`tincan.activity.Activity`
        :param agent: Agent object of desired state
        :type agent: :class:`tincan.agent.Agent`
        :param state_id: UUID of desired state
        :type state_id: str | unicode
        :param registration: registration UUID of desired state
        :type registration: str | unicode
        :return: LRS Response object with retrieved state document as content
        :rtype: :class:`tincan.lrs_response.LRSResponse`
        """
        if not isinstance(activity, Activity):
            activity = Activity(activity)

        if not isinstance(agent, Agent):
            agent = Agent(agent)

        request = HTTPRequest(
            method="GET",
            resource="activities/state",
            ignore404=True
        )

        request.query_params = {
            "activityId": activity.id,
            "agent": agent.to_json(self.version),
            "stateId": state_id
        }

        if registration is not None:
            request.query_params["registration"] = registration

        lrs_response = self._send_request(request)

        if lrs_response.success:
            doc = StateDocument(
                id=state_id,
                content=lrs_response.data,
                activity=activity,
                agent=agent
            )
            if registration is not None:
                doc.registration = registration

            headers = lrs_response.response.getheaders()
            if "lastModified" in headers and headers["lastModified"] is not None:
                doc.timestamp = headers["lastModified"]
            if "contentType" in headers and headers["contentType"] is not None:
                doc.content_type = headers["contentType"]
            if "etag" in headers and headers["etag"] is not None:
                doc.etag = headers["etag"]

            lrs_response.content = doc

        return lrs_response

    def save_state(self, state):
        """Save a state doc to the LRS

        :param state: State document to be saved
        :type state: :class:`tincan.documents.state_document.StateDocument`
        :return: LRS Response object with saved state as content
        :rtype: :class:`tincan.lrs_response.LRSResponse`
        """
        request = HTTPRequest(
            method="PUT",
            resource="activities/state",
            content=state.content,
        )
        if state.content_type is not None:
            request.headers["Content-Type"] = state.content_type
        else:
            request.headers["Content-Type"] = "application/octet-stream"

        if state.etag is not None:
            request.headers["If-Match"] = state.etag

        request.query_params = {
            "stateId": state.id,
            "activityId": state.activity.id,
            "agent": state.agent.to_json(self.version)
        }
        lrs_response = self._send_request(request)
        lrs_response.content = state

        return self._send_request(request)

    def _delete_state(self, activity, agent, state_id=None, registration=None, etag=None):
        """Private method to delete a specified state from the LRS

        :param activity: Activity object of state to be deleted
        :type activity: :class:`tincan.activity.Activity`
        :param agent: Agent object of state to be deleted
        :type agent: :class:`tincan.agent.Agent`
        :param state_id: UUID of state to be deleted
        :type state_id: str | unicode
        :param registration: registration UUID of state to be deleted
        :type registration: str | unicode
        :param etag: etag of state to be deleted
        :type etag: str | unicode
        :return: LRS Response object with deleted state as content
        :rtype: :class:`tincan.lrs_response.LRSResponse`
        """
        if not isinstance(activity, Activity):
            activity = Activity(activity)

        if not isinstance(agent, Agent):
            agent = Agent(agent)

        request = HTTPRequest(
            method="DELETE",
            resource="activities/state"
        )

        if etag is not None:
            request.headers["If-Match"] = etag

        request.query_params = {
            "activityId": activity.id,
            "agent": agent.to_json(self.version)
        }
        if state_id is not None:
            request.query_params["stateId"] = state_id

        if registration is not None:
            request.query_params["registration"] = registration

        lrs_response = self._send_request(request)

        return lrs_response

    def delete_state(self, state):
        """Delete a specified state from the LRS

        :param state: State document to be deleted
        :type state: :class:`tincan.documents.state_document.StateDocument`
        :return: LRS Response object
        :rtype: :class:`tincan.lrs_response.LRSResponse`
        """
        return self._delete_state(
            activity=state.activity,
            agent=state.agent,
            state_id=state.id,
            etag=state.etag
        )

    def clear_state(self, activity, agent, registration=None):
        """Clear state(s) with specified activity and agent

        :param activity: Activity object of state(s) to be deleted
        :type activity: :class:`tincan.activity.Activity`
        :param agent: Agent object of state(s) to be deleted
        :type agent: :class:`tincan.agent.Agent`
        :param registration: registration UUID of state(s) to be deleted
        :type registration: str | unicode
        :return: LRS Response object
        :rtype: :class:`tincan.lrs_response.LRSResponse`
        """
        return self._delete_state(
            activity=activity,
            agent=agent,
            registration=registration
        )

    def retrieve_activity_profile_ids(self, activity, since=None):
        """Retrieve activity profile id(s) with the specified parameters

        :param activity: Activity object of desired activity profiles
        :type activity: :class:`tincan.activity.Activity`
        :param since: Retrieve activity profile id's since this time
        :type since: str | unicode
        :return: LRS Response object with list of retrieved activity profile id's as content
        :rtype: :class:`tincan.lrs_response.LRSResponse`
        """
        if not isinstance(activity, Activity):
            activity = Activity(activity)

        request = HTTPRequest(
            method="GET",
            resource="activities/profile"
        )
        request.query_params["activityId"] = activity.id

        if since is not None:
            request.query_params["since"] = since

        lrs_response = self._send_request(request)

        if lrs_response.success:
            lrs_response.content = json.loads(lrs_response.data)

        return lrs_response

    def retrieve_activity_profile(self, activity, profile_id):
        """Retrieve activity profile with the specified parameters

        :param activity: Activity object of the desired activity profile
        :type activity: :class:`tincan.activity.Activity`
        :param profile_id: UUID of the desired profile
        :type profile_id: str | unicode
        :return: LRS Response object with an activity profile doc as content
        :rtype: :class:`tincan.lrs_response.LRSResponse`
        """
        if not isinstance(activity, Activity):
            activity = Activity(activity)

        request = HTTPRequest(
            method="GET",
            resource="activities/profile",
            ignore404=True
        )
        request.query_params = {
            "profileId": profile_id,
            "activityId": activity.id
        }
        lrs_response = self._send_request(request)

        if lrs_response.success:
            doc = ActivityProfileDocument(
                id=profile_id,
                content=lrs_response.data,
                activity=activity
            )
            headers = lrs_response.response.getheaders()
            if "lastModified" in headers and headers["lastModified"] is not None:
                doc.timestamp = headers["lastModified"]
            if "contentType" in headers and headers["contentType"] is not None:
                doc.content_type = headers["contentType"]
            if "etag" in headers and headers["etag"] is not None:
                doc.etag = headers["etag"]

            lrs_response.content = doc

        return lrs_response

    def save_activity_profile(self, profile):
        """Save an activity profile doc to the LRS

        :param profile: Activity profile doc to be saved
        :type profile: :class:`tincan.documents.activity_profile_document.ActivityProfileDocument`
        :return: LRS Response object with the saved activity profile doc as content
        :rtype: :class:`tincan.lrs_response.LRSResponse`
        """
        request = HTTPRequest(
            method="PUT",
            resource="activities/profile",
            content=profile.content
        )
        if profile.content_type is not None:
            request.headers["Content-Type"] = profile.content_type
        else:
            request.headers["Content-Type"] = "application/octet-stream"

        if profile.etag is not None:
            request.headers["If-Match"] = profile.etag

        request.query_params = {
            "profileId": profile.id,
            "activityId": profile.activity.id
        }
        lrs_response = self._send_request(request)
        lrs_response.content = profile

        return lrs_response

    def delete_activity_profile(self, profile):
        """Delete activity profile doc from LRS

        :param profile: Activity profile document to be deleted
        :type profile: :class:`tincan.documents.activity_profile_document.ActivityProfileDocument`
        :return: LRS Response object
        :rtype: :class:`tincan.lrs_response.LRSResponse`
        """
        request = HTTPRequest(
            method="DELETE",
            resource="activities/profile"
        )
        request.query_params = {
            "profileId": profile.id,
            "activityId": profile.activity.id
        }

        if profile.etag is not None:
            request.headers["If-Match"] = profile.etag

        return self._send_request(request)

    def retrieve_agent_profile_ids(self, agent, since=None):
        """Retrieve agent profile id(s) with the specified parameters

        :param agent: Agent object of desired agent profiles
        :type agent: :class:`tincan.agent.Agent`
        :param since: Retrieve agent profile id's since this time
        :type since: str | unicode
        :return: LRS Response object with list of retrieved agent profile id's as content
        :rtype: :class:`tincan.lrs_response.LRSResponse`
        """
        if not isinstance(agent, Agent):
            agent = Agent(agent)

        request = HTTPRequest(
            method="GET",
            resource="agents/profile"
        )
        request.query_params["agent"] = agent.to_json(self.version)

        if since is not None:
            request.query_params["since"] = since

        lrs_response = self._send_request(request)

        if lrs_response.success:
            lrs_response.content = json.loads(lrs_response.data)

        return lrs_response

    def retrieve_agent_profile(self, agent, profile_id):
        """Retrieve agent profile with the specified parameters

        :param agent: Agent object of the desired agent profile
        :type agent: :class:`tincan.agent.Agent`
        :param profile_id: UUID of the desired agent profile
        :type profile_id: str | unicode
        :return: LRS Response object with an agent profile doc as content
        :rtype: :class:`tincan.lrs_response.LRSResponse`
        """
        if not isinstance(agent, Agent):
            agent = Agent(agent)

        request = HTTPRequest(
            method="GET",
            resource="agents/profile",
            ignore404=True
        )
        request.query_params = {
            "profileId": profile_id,
            "agent": agent.to_json(self.version)
        }

        lrs_response = self._send_request(request)

        if lrs_response.success:
            doc = AgentProfileDocument(
                id=profile_id,
                content=lrs_response.data,
                agent=agent
            )
            headers = lrs_response.response.getheaders()
            if "lastModified" in headers and headers["lastModified"] is not None:
                doc.timestamp = headers["lastModified"]
            if "contentType" in headers and headers["contentType"] is not None:
                doc.content_type = headers["contentType"]
            if "etag" in headers and headers["etag"] is not None:
                doc.etag = headers["etag"]

            lrs_response.content = doc

        return lrs_response

    def save_agent_profile(self, profile):
        """Save an agent profile doc to the LRS

        :param profile: Agent profile doc to be saved
        :type profile: :class:`tincan.documents.agent_profile_document.AgentProfileDocument`
        :return: LRS Response object with the saved agent profile doc as content
        :rtype: :class:`tincan.lrs_response.LRSResponse`
        """
        request = HTTPRequest(
            method="PUT",
            resource="agents/profile",
            content=profile.content,
        )
        if profile.content_type is not None:
            request.headers["Content-Type"] = profile.content_type
        else:
            request.headers["Content-Type"] = "application/octet-stream"

        if profile.etag is not None:
            request.headers["If-Match"] = profile.etag

        request.query_params = {
            "profileId": profile.id,
            "agent": profile.agent.to_json(self.version)
        }
        lrs_response = self._send_request(request)
        lrs_response.content = profile

        return lrs_response

    def delete_agent_profile(self, profile):
        """Delete agent profile doc from LRS

        :param profile: Agent profile document to be deleted
        :type profile: :class:`tincan.documents.agent_profile_document.AgentProfileDocument`
        :return: LRS Response object
        :rtype: :class:`tincan.lrs_response.LRSResponse`
        """
        request = HTTPRequest(
            method="DELETE",
            resource="agents/profile"
        )
        request.query_params = {
            "profileId": profile.id,
            "agent": profile.agent.to_json(self.version)
        }

        if profile.etag is not None:
            request.headers["If-Match"] = profile.etag

        return self._send_request(request)

    @property
    def endpoint(self):
        """The endpoint of the Remote LRS

        :setter: Tries to convert to unicode.  Appends a "/" if necessary. Prepends http:// if necessary.
        :setter type: str | unicode
        :rtype: unicode
        """
        return self._endpoint

    @endpoint.setter
    def endpoint(self, value):
        if value is not None:
            if not isinstance(value, str):
                value = str(value)
            if not value.endswith("/"):
                value += "/"
            if not value.startswith("http"):
                value = "http://" + value

        self._endpoint = value

    @property
    def version(self):
        """Version being used for remote LRS communication

        :setter: Tries to convert to unicode.  Must be a supported
        version. Setting to None defaults to the latest version.
        :setter type: str | unicode
        :rtype: unicode
        """
        return self._version

    @version.setter
    def version(self, value):
        if value is not None:
            if not isinstance(value, str):
                str(value)
            if value not in Version.supported:
                raise Exception("Unsupported Version")
        else:
            value = Version.latest

        self._version = value

    @property
    def auth(self):
        """Authority being used for remote LRS communication

        :setter: Tries to convert to unicode.
        :setter type: str | unicode
        :rtype: unicode
        """
        return self._auth

    @auth.setter
    def auth(self, value):
        if value is not None and not isinstance(value, str):
            str(value)
        self._auth = value

    def get_endpoint_server_root(self):
        """Parses RemoteLRS object's endpoint and returns its root

        :return: Root of the RemoteLRS object endpoint
        :rtype: unicode
        """
        parsed = urlparse(self._endpoint)
        root = parsed.scheme + "://" + parsed.hostname

        if parsed.port is not None:
            root += ":" + str(parsed.port)

        return root



# Copyright 2014 Rustici Software
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from tincan.serializable_base import SerializableBase


class Extensions(dict, SerializableBase):
    """
    Contains domain-specific custom data.

    Can be created from a dict, another :class:`tincan.Extensions`,
    or from args and kwargs.

    Use this like a regular Python dict.
    """

    def __init__(self, *args, **kwargs):
        super(Extensions, self).__init__(*args, **kwargs)



# Copyright 2014 Rustici Software
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import uuid

from tincan.serializable_base import SerializableBase


"""
.. module StatementRef
   :synopsis: A StatementRef object that is a reference to another pre-existing statement.
"""


class StatementRef(SerializableBase):
    _props_req = [
        'object_type'
    ]

    _props = [
        'id'
    ]

    _props.extend(_props_req)

    def __init__(self, *args, **kwargs):
        self._object_type = None
        self._id = None

        super(StatementRef, self).__init__(*args, **kwargs)

    @property
    def object_type(self):
        """Object type for Statement Ref. Will always be "StatementRef"

        :setter: Tries to convert to unicode
        :setter type: unicode
        :rtype: unicode

        """
        return self._object_type

    @object_type.setter
    def object_type(self, _):
        self._object_type = 'StatementRef'

    @property
    def id(self):
        """Id for Statement Ref

        :setter: Tries to convert to unicode
        :setter type: unicode
        :rtype: unicode

        """
        return self._id

    @id.setter
    def id(self, value):
        if value is not None and not isinstance(value, uuid.UUID):
            if isinstance(value, str) and not self._UUID_REGEX.match(value):
                raise ValueError("Invalid UUID string")
            value = uuid.UUID(value)
        self._id = value

    @id.deleter
    def id(self):
        del self._id



# Copyright 2014 Rustici Software
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from tincan.agent import Agent
from tincan.agent_list import AgentList

"""

.. module:: Group
   :synopsis: An object that contains a group of Agents

"""


class Group(Agent):
    _props = [
        "member"
    ]

    _props.extend(Agent._props)

    def __init__(self, *args, **kwargs):
        self._object_type = None
        self._member = AgentList()

        super(Group, self).__init__(*args, **kwargs)

    def addmember(self, value):
        """Adds a single member to this group's list of members.
        Tries to convert to :class:`tincan.Agent`

        :param value: The member to add to this group
        :type value: :class:`tincan.Agent`

        """

        if value is not None and not isinstance(value, Agent):
            value = Agent(value)

        self._member.append(value)

    @property
    def member(self):
        """Members for Group

        :setter: Tries to convert to :class:`tincan.AgentList`
        :setter type: :class:`tincan.AgentList`
        :rtype: :class:`tincan.AgentList`
        """
        return self._member

    @member.setter
    def member(self, value):
        newmembers = AgentList()
        if value is not None:
            if isinstance(value, list):
                for k in value:
                    if not isinstance(k, Agent):
                        newmembers.append(Agent(k))
                    else:
                        newmembers.append(k)
            else:
                newmembers = AgentList(value)
        self._member = newmembers

    @member.deleter
    def member(self):
        del self._member

    @property
    def object_type(self):
        """Object type for Group. Will always be "Group"

        :setter: Tries to convert to unicode
        :setter type: unicode
        :rtype: unicode

        """
        return self._object_type

    @object_type.setter
    def object_type(self, _):
        self._object_type = 'Group'



# Copyright 2014 Rustici Software
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from tincan.base import Base


class HTTPRequest(Base):
    """Creates a new HTTPRequest object, either from a dict, another object, or from kwargs

    :param method: Method for the HTTP connection ("GET", "POST", "DELETE", etc.)
    :type method: unicode
    :param resource: Resource for the LRS HTTP connection ("about", "statements", "activities/state", etc.)
    :type resource: unicode
    :param headers: Headers for the HTTP connection ("If-Match", "Content-Type", etc.)
    :type headers: dict(unicode:unicode)
    :param query_params: Query parameters for the HTTP connection ("registration", "since", "statementId", etc.)
    :type query_params: dict(unicode:unicode)
    :param content: Content body for the HTTP connection. Valid json string.
    :type content: unicode
    :param ignore404: True if this request should consider a 404 response successful, False otherwise
    :type ignore404: bool
    """

    _props_req = [
        'method',
        'resource',
        'headers',
        'query_params',
    ]

    _props = [
        'content',
        'ignore404',
    ]

    _props.extend(_props_req)

    def __init__(self, *args, **kwargs):
        self._method = None
        self._resource = None
        self._headers = None
        self._query_params = None
        self._content = None
        self._ignore404 = None

        super(HTTPRequest, self).__init__(*args, **kwargs)

    @property
    def method(self):
        """Method for the HTTP connection ("GET", "POST", "DELETE", etc.)

        :setter: Tries to convert to unicode
        :setter type: str | unicode
        :rtype: unicode
        """
        return self._method

    @method.setter
    def method(self, value):
        if not isinstance(value, str) and value is not None:
            str(value)
        self._method = value

    @property
    def resource(self):
        """Resource for the LRS HTTP connection ("about", "statements", "activities/state", etc.)

        :setter: Tries to convert to unicode
        :setter type: str | unicode
        :rtype: unicode
        """
        return self._resource

    @resource.setter
    def resource(self, value):
        if not isinstance(value, str) and value is not None:
            str(value)
        self._resource = value

    @property
    def headers(self):
        """Headers for the HTTP connection ("If-Match", "Content-Type", etc.)

        :setter: Accepts a dict or an object
        :setter type: dict
        :rtype: dict
        """
        return self._headers

    @headers.setter
    def headers(self, value):
        val_dict = {}
        if value is not None:
            val_dict.update(value if isinstance(value, dict) else vars(value))
        self._headers = val_dict

    @property
    def query_params(self):
        """Query parameters for the HTTP connection ("registration", "since", "statementId", etc.)

        :setter: Accepts a dict or an object
        :setter type: dict
        :rtype: dict
        """
        return self._query_params

    @query_params.setter
    def query_params(self, value):
        val_dict = {}
        if value is not None:
            val_dict.update(value if isinstance(value, dict) else vars(value))
        self._query_params = val_dict

    @property
    def content(self):
        """Content body for the HTTP connection. Valid json string.

        :setter: Tries to convert to unicode
        :setter type: str | unicode
        :rtype: unicode
        """
        return self._content

    @content.setter
    def content(self, value):
        if not isinstance(value, str) and value is not None:
            value = value.decode("utf-8")
        self._content = value

    @content.deleter
    def content(self):
        del self._content

    @property
    def ignore404(self):
        """True if this request should consider a 404 response successful, False otherwise

        :setter: Tries to convert to boolean
        :setter type: bool
        :rtype: bool
        """
        return self._ignore404

    @ignore404.setter
    def ignore404(self, value):
        self._ignore404 = bool(value)

    @ignore404.deleter
    def ignore404(self):
        del self._ignore404



# Copyright 2014 Rustici Software
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from tincan.serializable_base import SerializableBase
from tincan.language_map import LanguageMap

"""
.. module:: Attachment
   :synopsis: An Attachment object that contains an attached file and its information

"""


class Attachment(SerializableBase):
    _props_req = [
        "usage_type",
        "display",
        "content_type",
        "length",
        "sha2"
    ]

    _props = [
        "description",
        "fileurl"
    ]

    _props.extend(_props_req)

    def __init__(self, *args, **kwargs):
        self._usage_type = None
        self._display = None
        self._content_type = None
        self._length = None
        self._sha2 = None
        self._description = None
        self._fileurl = None

        super(Attachment, self).__init__(*args, **kwargs)

    @property
    def usage_type(self):
        """Usage type for Attachment

        :setter: Tries to convert to unicode
        :setter type: unicode
        :rtype: unicode

        """
        return self._usage_type

    @usage_type.setter
    def usage_type(self, value):
        if value is not None:
            if value == '':
                raise ValueError("Property usage_type can not be set to an empty string")
            elif not isinstance(value, str):
                value = str(value)
        self._usage_type = value

    @usage_type.deleter
    def usage_type(self):
        del self._usage_type

    @property
    def content_type(self):
        """Content type for Attachment

        :setter: Tries to convert to unicode
        :setter type: unicode
        :rtype: unicode

        """
        return self._content_type

    @content_type.setter
    def content_type(self, value):
        if value is not None:
            if value == '':
                raise ValueError("Property content_type can not be set to an empty string")
            elif not isinstance(value, str):
                value = str(value)
        self._content_type = value

    @content_type.deleter
    def content_type(self):
        del self._content_type

    @property
    def length(self):
        """Usage type for Attachment

        :setter: Tries to convert to long
        :setter type: int | long
        :rtype: long

        """
        return self._length

    @length.setter
    def length(self, value):
        if value is not None:
            if not isinstance(value, int):
                value = int(value)
        self._length = value

    @length.deleter
    def length(self):
        del self._length

    @property
    def sha2(self):
        """Sha2 for Attachment

        :setter: Tries to convert to unicode
        :setter type: unicode
        :rtype: unicode

        """
        return self._sha2

    @sha2.setter
    def sha2(self, value):
        if value is not None:
            if value == '':
                raise ValueError("Property sha2 can not be set to an empty string")
            elif not isinstance(value, str):
                value = str(value)
        self._sha2 = value

    @sha2.deleter
    def sha2(self):
        del self._sha2

    @property
    def fileurl(self):
        """File URL for Attachment

        :setter: Tries to convert to unicode
        :setter type: unicode
        :rtype: unicode

        """
        return self._fileurl

    @fileurl.setter
    def fileurl(self, value):
        if value is not None:
            if value == '':
                raise ValueError("Property fileurl can not be set to an empty string")
            elif not isinstance(value, str):
                value = str(value)
        self._fileurl = value

    @fileurl.deleter
    def fileurl(self):
        del self._fileurl

    @property
    def display(self):
        """Display for Attachment

        :setter: Tries to convert to :class:`tincan.LanguageMap`
        :setter type: :class:`tincan.LanguageMap`
        :rtype: :class:`tincan.LanguageMap`

        """
        return self._display

    @display.setter
    def display(self, value):
        if value is not None and not isinstance(value, LanguageMap):
            value = LanguageMap(value)
        self._display = value

    @display.deleter
    def display(self):
        del self._display

    @property
    def description(self):
        """Description for Attachment

        :setter: Tries to convert to :class:`tincan.LanguageMap`
        :setter type: :class:`tincan.LanguageMap`
        :rtype: :class:`tincan.LanguageMap`

        """
        return self._description

    @description.setter
    def description(self, value):
        if value is not None and not isinstance(value, LanguageMap):
            value = LanguageMap(value)
        self._description = value

    @description.deleter
    def description(self):
        del self._description



# Copyright 2014 Rustici Software
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from tincan.serializable_base import SerializableBase

"""

.. module:: Agent Account
   :synopsis: The account object that can be included in an agent

"""


class AgentAccount(SerializableBase):
    _props = [
        "name",
        "home_page"
    ]

    def __init__(self, *args, **kwargs):
        self._name = None
        self._home_page = None

        super(AgentAccount, self).__init__(*args, **kwargs)

    @property
    def name(self):
        """Name for Account

        :setter: Tries to convert to unicode
        :setter type: unicode
        :rtype: unicode

        """
        return self._name

    @name.setter
    def name(self, value):
        if value is not None:
            if value == '':
                raise ValueError("Property name can not be set to an empty string")
            elif not isinstance(value, str):
                value = str(value)
        self._name = value

    @name.deleter
    def name(self):
        del self._name

    @property
    def home_page(self):
        """Homepage for Account

        :setter: Tries to convert to unicode
        :setter type: unicode
        :rtype: unicode

        """
        return self._home_page

    @home_page.setter
    def home_page(self, value):
        if value is not None:
            if value == '':
                raise ValueError("Property homepage can not be set to an empty string")
            elif not isinstance(value, str):
                value = str(value)
        self._home_page = value

    @home_page.deleter
    def home_page(self):
        del self._home_page



# Copyright 2014 Rustici Software
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from tincan.serializable_base import SerializableBase
from tincan.agent_account import AgentAccount

"""

.. module:: Agent
   :synopsis: The Agent object that contains the information about an Agent

"""


class Agent(SerializableBase):
    _props_req = [
        "object_type"
    ]

    _props = [
        "name",
        "mbox",
        "mbox_sha1sum",
        "openid",
        "account"
    ]

    _props.extend(_props_req)

    def __init__(self, *args, **kwargs):
        self._object_type = None
        self._name = None
        self._mbox = None
        self._mbox_sha1sum = None
        self._openid = None
        self._account = None

        super(Agent, self).__init__(*args, **kwargs)

    @property
    def object_type(self):
        """Object Type for Agent. Will always be 'Agent'

        :setter: Tries to convert to unicode
        :setter type: unicode
        :rtype: unicode

        """
        return self._object_type

    @object_type.setter
    def object_type(self, _):
        self._object_type = 'Agent'

    @property
    def name(self):
        """Name for Agent

        :setter: Tries to convert to unicode
        :setter type: unicode
        :rtype: unicode

        """
        return self._name

    @name.setter
    def name(self, value):
        if value is not None:
            if value == '':
                raise ValueError("Property name can not be set to an empty string")
            elif not isinstance(value, str):
                value = str(value)
        self._name = value

    @name.deleter
    def name(self):
        del self._name

    @property
    def mbox(self):
        """Mbox for Agent

        :setter: Tries to convert to unicode
        :setter type: unicode
        :rtype: unicode

        """
        return self._mbox

    @mbox.setter
    def mbox(self, value):
        if value is not None:
            if value == '':
                raise ValueError("Property mbox can not be set to an empty string")
            elif not isinstance(value, str):
                value = str(value)
        if not value.startswith("mailto:"):
            value = "mailto:" + value
        self._mbox = value

    @mbox.deleter
    def mbox(self):
        del self._mbox

    @property
    def mbox_sha1sum(self):
        """Mbox_sha1sum for Agent

        :setter: Tries to convert to unicode
        :setter type: unicode
        :rtype: unicode

        """
        return self._mbox_sha1sum

    @mbox_sha1sum.setter
    def mbox_sha1sum(self, value):
        if value is not None:
            if value == '':
                raise ValueError("Property mbox_sha1sum can not be set to an empty string")
            elif not isinstance(value, str):
                value = str(value)
        self._mbox_sha1sum = value

    @mbox_sha1sum.deleter
    def mbox_sha1sum(self):
        del self._mbox_sha1sum

    @property
    def openid(self):
        """Openid for Agent

        :setter: Tries to convert to unicode
        :setter type: unicode
        :rtype: unicode

        """
        return self._openid

    @openid.setter
    def openid(self, value):
        if value is not None:
            if value == '':
                raise ValueError("Property openid can not be set to an empty string")
            elif not isinstance(value, str):
                value = str(value)
        self._openid = value

    @openid.deleter
    def openid(self):
        del self._openid

    @property
    def account(self):
        """Account for Agent

        :setter: Tries to convert to :class:`tincan.AgentAccount`
        :setter type: :class:`tincan.AgentAccount`
        :rtype: :class:`tincan.AgentAccount`

        """
        return self._account

    @account.setter
    def account(self, value):
        if value is not None and not isinstance(value, AgentAccount):
            value = AgentAccount(value)
        self._account = value

    @account.deleter
    def account(self):
        del self._account



# Copyright 2014 Rustici Software
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import uuid
import re

from tincan.serializable_base import SerializableBase
from tincan.group import Group
from tincan.agent import Agent
from tincan.extensions import Extensions
from tincan.context_activities import ContextActivities
from tincan.statement_ref import StatementRef


"""
.. module:: context
   :synopsis: A context object that provides contextual information
   about a statement

"""


class Context(SerializableBase):
    _LANG_REGEX = re.compile(
        '^(((([A-Za-z]{2,3}(-([A-Za-z]{3}(-[A-Za-z]{3}){0,2}))?)|[A-Za-z]{4}|[A-Za-z]{5,8})(-([A-Za-z]{4}))?(-([A-Za-z]'
        '{2}|[0-9]{3}))?(-([A-Za-z0-9]{5,8}|[0-9][A-Za-z0-9]{3}))*(-([0-9A-WY-Za-wy-z](-[A-Za-z0-9]{2,8})+))*(-(x(-[A-Z'
        'a-z0-9]{1,8})+))?)|(x(-[A-Za-z0-9]{1,8})+)|((en-GB-oed|i-ami|i-bnn|i-default|i-enochian|i-hak|i-klingon|i-lux|'
        'i-mingo|i-navajo|i-pwn|i-tao|i-tay|i-tsu|sgn-BE-FR|sgn-BE-NL|sgn-CH-DE)|(art-lojban|cel-gaulish|no-bok|no-nyn|'
        'zh-guoyu|zh-hakka|zh-min|zh-min-nan|zh-xiang)))$')

    _props = [
        'registration',
        'instructor',
        'team',
        'context_activities',
        'revision',
        'platform',
        'language',
        'statement',
        'extensions'
    ]

    def __init__(self, *args, **kwargs):
        self._registration = None
        self._instructor = None
        self._team = None
        self._context_activities = None
        self._revision = None
        self._platform = None
        self._language = None
        self._statement = None
        self._extensions = None

        super(Context, self).__init__(*args, **kwargs)

    @property
    def registration(self):
        """Registration for Context

        :setter: Tries to convert to unicode
        :setter type: unicode
        :rtype: unicode

        """
        return self._registration

    @registration.setter
    def registration(self, value):
        if value is not None and not isinstance(value, uuid.UUID):
            if isinstance(value, str) and not self._UUID_REGEX.match(value):
                raise ValueError("Invalid UUID string")
            value = uuid.UUID(value)
        self._registration = value

    @registration.deleter
    def registration(self):
        del self._registration

    @property
    def instructor(self):
        """Instructor for Context

        :setter: Tries to convert to :class:`tincan.Agent` or :class:`tincan.Group`
        :setter type: :class:`tincan.Agent` | :class:`tincan.Group`
        :rtype: :class:`tincan.Agent` | :class:`tincan.Group`

        """
        return self._instructor

    @instructor.setter
    def instructor(self, value):
        if value is not None and not isinstance(value, Group) and not isinstance(value, Agent):
            try:
                value = Agent(value)
            except (TypeError, AttributeError):
                value = Group(value)
        self._instructor = value

    @instructor.deleter
    def instructor(self):
        del self._instructor

    @property
    def team(self):
        """Team for Context

        :setter: Tries to convert to :class:`tincan.Group`
        :setter type: :class:`tincan.Group`
        :rtype: :class:`tincan.Group`

        """
        return self._team

    @team.setter
    def team(self, value):
        if value is not None and not isinstance(value, Group):
            value = Group(value)
        self._team = value

    @team.deleter
    def team(self):
        del self._team

    @property
    def context_activities(self):
        """Context Activities for Context

        :setter: Tries to convert to :class:`tincan.ContextActivities`
        :setter type: :class:`tincan.ContextActivities`
        :rtype: :class:`tincan.ContextActivities`

        """
        return self._context_activities

    @context_activities.setter
    def context_activities(self, value):
        if value is not None and not isinstance(value, ContextActivities):
            value = ContextActivities(value)
        self._context_activities = value

    @context_activities.deleter
    def context_activities(self):
        del self._context_activities

    @property
    def revision(self):
        """Revision for Context

        :setter: Tries to convert to unicode
        :setter type: unicode
        :rtype: unicode

        """
        return self._revision

    @revision.setter
    def revision(self, value):
        if value is not None and not isinstance(value, str):
            value = str(value)
        self._revision = value

    @revision.deleter
    def revision(self):
        del self._revision

    @property
    def platform(self):
        """Platform for Context

        :setter: Tries to convert to unicode
        :setter type: unicode
        :rtype: unicode

        """
        return self._platform

    @platform.setter
    def platform(self, value):
        if value is not None and not isinstance(value, str):
            value = str(value)
        self._platform = value

    @platform.deleter
    def platform(self):
        del self._platform

    @property
    def language(self):
        """Language for Context

        :setter: Tries to convert to unicode
        :setter type: unicode
        :rtype: unicode

        """
        return self._language

    @language.setter
    def language(self, value):
        if value is not None:
            if not isinstance(value, str):
                value = str(value)
            if not self._LANG_REGEX.match(value):
                raise ValueError("invalid regional identifier")
        self._language = value

    @language.deleter
    def language(self):
        del self._language

    @property
    def statement(self):
        """Statement for Context

        :setter: Tries to convert to :class:`tincan.StatementRef`
        :setter type: :class:`tincan.StatementRef`
        :rtype: :class:`tincan.StatementRef`

        """
        return self._statement

    @statement.setter
    def statement(self, value):
        if value is not None and not isinstance(value, StatementRef):
            value = StatementRef(value)
        self._statement = value

    @statement.deleter
    def statement(self):
        del self._statement

    @property
    def extensions(self):
        """Extensions for Context

        :setter: Tries to convert to :class:`tincan.Extensions`
        :setter type: :class:`tincan.Extensions`
        :rtype: :class:`tincan.Extensions`

        """
        return self._extensions

    @extensions.setter
    def extensions(self, value):
        if value is not None and not isinstance(value, Extensions):
            value = Extensions(value)
        self._extensions = value

    @extensions.deleter
    def extensions(self):
        del self._extensions



# Copyright 2014 Rustici Software
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import json
import uuid
import datetime
import re

from tincan.base import Base
from tincan.version import Version
from tincan.conversions.iso8601 import jsonify_datetime, jsonify_timedelta


"""
.. module:: serializable_base
   :synopsis: A base object that provides the common initializer from :class:`tincan.Base`
   as well as common serialization functionality

"""


class SerializableBase(Base):
    _props_corrected = {
        '_more_info': 'moreInfo',
        '_interaction_type': 'interactionType',
        '_correct_responses_pattern': 'correctResponsesPattern',
        '_object_type': 'objectType',
        '_usage_type': 'usageType',
        '_content_type': 'contentType',
        '_fileurl': 'fileUrl',
        '_context_activities': 'contextActivities',
        '_home_page': 'homePage',
    }

    _UUID_REGEX = re.compile(
        r'^[a-f0-9]{8}-'
        r'[a-f0-9]{4}-'
        r'[1-5][a-f0-9]{3}-'
        r'[89ab][a-f0-9]{3}-'
        r'[a-f0-9]{12}$'
    )

    def __init__(self, *args, **kwargs):

        new_kwargs = {}
        for obj in args:
            new_kwargs.update(obj if isinstance(obj, dict) else vars(obj))

        new_kwargs.update(kwargs)

        for uscore, camel in self._props_corrected.items():
            if camel in new_kwargs:
                new_kwargs[uscore[1:]] = new_kwargs[camel]
                new_kwargs.pop(camel)

        super(SerializableBase, self).__init__(**new_kwargs)

    @classmethod
    def from_json(cls, json_data):
        """Tries to convert a JSON representation to an object of the same
        type as self

        A class can provide a _fromJSON implementation in order to do specific
        type checking or other custom implementation details. This method
        will throw a ValueError for invalid JSON, a TypeError for
        improperly constructed, but valid JSON, and any custom errors
        that can be be propagated from class constructors.

        :param json_data: The JSON string to convert
        :type json_data: str | unicode

        :raises: TypeError, ValueError, LanguageMapInitError
        """

        data = json.loads(json_data)
        result = cls(data)
        if hasattr(result, "_from_json"):
            result._from_json()
        return result

    def to_json(self, version=Version.latest):
        """Tries to convert an object into a JSON representation and return
        the resulting string

        An Object can define how it is serialized by overriding the as_version()
        implementation. A caller may further define how the object is serialized
        by passing in a custom encoder. The default encoder will ignore
        properties of an object that are None at the time of serialization.

        :param version: The version to which the object must be serialized to.
        This will default to the latest version supported by the library.
        :type version: str | unicode

        """
        return json.dumps(self.as_version(version))

    def as_version(self, version=Version.latest):
        """Returns a dict that has been modified based on versioning
        in order to be represented in JSON properly

        A class should overload as_version(self, version)
        implementation in order to tailor a more specific representation

        :param version: the relevant version. This allows for variance
         between versions
        :type version: str | unicode

        """
        if not isinstance(self, list):
            result = {}
            for k, v in iter(self.items()) if isinstance(self, dict) else iter(vars(self).items()):
                k = self._props_corrected.get(k, k)
                if isinstance(v, SerializableBase):
                    result[k] = v.as_version(version)
                elif isinstance(v, list):
                    result[k] = []
                    for val in v:
                        if isinstance(val, SerializableBase):
                            result[k].append(val.as_version(version))
                        else:
                            result[k].append(val)
                elif isinstance(v, uuid.UUID):
                    result[k] = str(v)
                elif isinstance(v, datetime.timedelta):
                    result[k] = jsonify_timedelta(v)
                elif isinstance(v, datetime.datetime):
                    result[k] = jsonify_datetime(v)
                else:
                    result[k] = v
            result = self._filter_none(result)
        else:
            result = []
            for v in self:
                if isinstance(v, SerializableBase):
                    result.append(v.as_version(version))
                else:
                    result.append(v)
        return result

    @staticmethod
    def _filter_none(obj):
        """Filters out attributes set to None prior to serialization, and
        returns a new object without those attributes. This saves
        the serializer from sending empty bytes over the network. This method also
        fixes the keys to look as expected by ignoring a leading '_' if it
        is present.

        :param obj: the dictionary representation of an object that may have
        None attributes
        :type obj: dict

        """
        result = {}
        for k, v in obj.items():
            if v is not None:
                if k.startswith('_'):
                    k = k[1:]
                result[k] = v
        return result



from setuptools import setup

setup(
    name='tincan',
    packages=[
        'tincan',
        'tincan/conversions',
        'tincan/documents',
    ],
    version='1.0.0',
    description='A Python library for implementing Tin Can API.',
    author='Rustici Software',
    author_email='mailto:support+tincanpython@tincanapi.com',
    maintainer='Brian J. Miller',
    maintainer_email='mailto:brian.miller@tincanapi.com',
    url='http://rusticisoftware.github.io/TinCanPython/',
    license='Apache License 2.0',
    keywords=[
        'Tin Can',
        'TinCan',
        'Tin Can API',
        'TinCanAPI',
        'Experience API',
        'xAPI',
        'SCORM',
        'AICC',
    ],
    install_requires=[
        'aniso8601',
        'pytz',
    ],
)



# Copyright 2014 Rustici Software
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from tincan.serializable_base import SerializableBase
from tincan.statement_list import StatementList

"""
.. module:: statements_result
   :synopsis: Statements result model class, returned by LRS calls to get
              multiple statements.
"""


class StatementsResult(SerializableBase):
    _props_req = [
        'statements',
        'more',
    ]

    _props = []
    _props.extend(_props_req)

    def __init__(self, *args, **kwargs):
        self._statements = None
        self._more = None

        super(StatementsResult, self).__init__(*args, **kwargs)

    @property
    def statements(self):
        """Statements for StatementsResult

        :setter: Tries to convert each element to :class:`tincan.Statement`
        :setter type: list of :class:`tincan.Statement`
        :rtype: list of :class:`tincan.Statement`

        """
        return self._statements

    @statements.setter
    def statements(self, value):
        if value is None:
            self._statements = StatementList()
            return
        try:
            self._statements = StatementList(value)
        except Exception:
            raise TypeError(f"Property 'statements' in a 'tincan.{self.__class__.__name__}' object must be set with a "
                            f"list or None."
                            f"\n\n"
                            f"Tried to set it with a '{value.__class__.__name__}' object: {repr(value)}"
                            f"\n\n")


    @property
    def more(self):
        """More for StatementsResult

        :setter: Tries to convert to unicode
        :setter type: unicode
        :rtype: unicode

        """
        return self._more

    @more.setter
    def more(self, value):
        if value is None or isinstance(value, str):
            self._more = value
            return
        try:
            self._more = str(value)
        except Exception as e:
            msg = (
                f"Property 'more' in a 'tincan.{self.__class__.__name__}' object must be set with a "
                f"str or None."
            )
            msg += repr(e)
            raise TypeError(msg) from e



from tincan.documents.document import Document
from tincan.documents.state_document import StateDocument
from tincan.documents.activity_profile_document import ActivityProfileDocument
from tincan.documents.agent_profile_document import AgentProfileDocument



# Copyright 2014 Rustici Software
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from datetime import datetime

from tincan.serializable_base import SerializableBase
from tincan.agent import Agent
from tincan.group import Group
from tincan.verb import Verb
from tincan.context import Context
from tincan.attachment import Attachment
from tincan.attachment_list import AttachmentList
from tincan.conversions.iso8601 import make_datetime


"""

.. module:: StatementBase
   :synopsis: The base object for both Statement and SubStatement

"""


class StatementBase(SerializableBase):
    _props_req = [
        'actor',
        'verb',
        'object',
        'timestamp',
    ]

    _props = [
        'context',
        'attachments'
    ]

    _props.extend(_props_req)

    def __init__(self, *args, **kwargs):
        self._actor = None
        self._verb = None
        self._object = None
        self._timestamp = None
        self._context = None
        self._attachments = None

        super(StatementBase, self).__init__(*args, **kwargs)

    @property
    def actor(self):
        """Actor for StatementBase

        :setter: Tries to convert to :class:`tincan.Agent` or :class:`tincan.Group`
        :setter type: :class:`tincan.Agent` | :class:`tincan.Group`
        :rtype: :class:`tincan.Agent` | :class:`tincan.Group`

        """
        return self._actor

    @actor.setter
    def actor(self, value):
        if value is not None and not isinstance(value, Agent) and not isinstance(value, Group):
            if isinstance(value, dict):
                if 'object_type' in value or 'objectType' in value:
                    if 'objectType' in value:
                        value['object_type'] = value['objectType']
                        value.pop('objectType')
                    if value['object_type'] == 'Agent':
                        value = Agent(value)
                    elif value['object_type'] == 'Group':
                        value = Group(value)
                    else:
                        value = Agent(value)
                else:
                    value = Agent(value)
        self._actor = value

    @actor.deleter
    def actor(self):
        del self._actor

    @property
    def verb(self):
        """Verb for StatementBase

        :setter: Tries to convert to :class:`tincan.Verb`
        :setter type: :class:`tincan.Verb`
        :rtype: :class:`tincan.Verb`

        """
        return self._verb

    @verb.setter
    def verb(self, value):
        if value is not None and not isinstance(value, Verb):
            value = Verb(value)
        self._verb = value

    @verb.deleter
    def verb(self):
        del self._verb

    @property
    def timestamp(self):
        """Timestamp for StatementBase

        :setter: Tries to convert to :class:`datetime.datetime`. If
        no timezone is given, makes a naive `datetime.datetime`.

        Strings will be parsed as ISO 8601 timestamps.

        If a number is provided, it will be interpreted as a UNIX
        timestamp, which by definition is UTC.

        If a `dict` is provided, does `datetime.datetime(**value)`.

        If a `tuple` or a `list` is provided, does
        `datetime.datetime(*value)`. Uses the timezone in the tuple or
        list if provided.

        :setter type: :class:`datetime.datetime` | unicode | str | int | float | dict | tuple | list | None
        :rtype: :class:`datetime.datetime`
        """
        return self._timestamp

    @timestamp.setter
    def timestamp(self, value):
        if value is None or isinstance(value, datetime):
            self._timestamp = value
            return

        try:
            self._timestamp = make_datetime(value)
        except TypeError as e:
            message = (
                f"Property 'timestamp' in a 'tincan.{self.__class__.__name__}' "
                f"object must be set with a "
                f"datetime.datetime, str, unicode, int, float, dict "
                f"or None.\n\n{repr(e)}"
            )
            raise TypeError(message) from e

    @timestamp.deleter
    def timestamp(self):
        del self._timestamp

    @property
    def context(self):
        """Context for StatementBase

        :setter: Tries to convert to :class:`tincan.Context`
        :setter type: :class:`tincan.Context`
        :rtype: :class:`tincan.Context`

        """
        return self._context

    @context.setter
    def context(self, value):
        if value is not None and not isinstance(value, Context):
            value = Context(value)
        self._context = value

    @context.deleter
    def context(self):
        del self._context

    @property
    def attachments(self):
        """Attachments for StatementBase

        :setter: Tries to convert each element to :class:`tincan.Attachment`
        :setter type: :class:`tincan.AttachmentList`
        :rtype: :class:`tincan.AttachmentList`

        """
        return self._attachments

    @attachments.setter
    def attachments(self, value):
        if value is not None and not isinstance(value, AttachmentList):
            try:
                value = AttachmentList([Attachment(value)])
            except (TypeError, AttributeError):
                value = AttachmentList(value)
        self._attachments = value

    @attachments.deleter
    def attachments(self):
        del self._attachments



# Copyright 2014 Rustici Software
#
# Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from tincan.statement_base import StatementBase
from tincan.agent import Agent
from tincan.group import Group
from tincan.activity import Activity


class SubStatement(StatementBase):
    _props_req = [
        'object_type'
    ]

    _props = []

    _props.extend(StatementBase._props)
    _props.extend(_props_req)

    def __init__(self, *args, **kwargs):
        self._object_type = None

        super(SubStatement, self).__init__(*args, **kwargs)

    @property
    def object(self):
        """Object for SubStatement

        :setter: Setter for object
        :setter type: :class:`tincan.Agent` | :class:`tincan.Group` | :class:`tincan.Activity`
        :rtype: :class:`tincan.Agent` | :class:`tincan.Group` | :class:`tincan.Activity`

        """
        return self._object

    @object.setter
    def object(self, value):
        if value is not None and \
                not isinstance(value, Agent) and \
                not isinstance(value, Group) and \
                not isinstance(value, Activity):
            if isinstance(value, dict):
                if 'object_type' in value or 'objectType' in value:
                    if 'objectType' in value:
                        value['object_type'] = value['objectType']
                        value.pop('objectType')
                    if value['object_type'] == 'Agent':
                        value = Agent(value)
                    elif value['object_type'] == 'Activity':
                        value = Activity(value)
                    elif value['object_type'] == 'Group':
                        value = Group(value)
                    else:
                        value = Activity(value)
                else:
                    value = Activity(value)
        self._object = value

    @object.deleter
    def object(self):
        del self._object

    @property
    def object_type(self):
        """Object Type for SubStatement. Will always be "SubStatement"

        :setter: Tries to convert to unicode
        :setter type: unicode
        :rtype: unicode

        """
        return self._object_type

    @object_type.setter
    def object_type(self, _):
        self._object_type = 'SubStatement'



# Copyright 2014 Rustici Software
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import datetime

from tincan.base import Base
from tincan.conversions.iso8601 import make_datetime


class Document(Base):
    """Document class can be instantiated from a dict, another :class:`tincan.Document`, or from kwargs

    :param id: The id of this document
    :type id: unicode
    :param content_type: The content type of the content of this document
    :type content_type: unicode
    :param content: The content of this document
    :type content: bytearray
    :param etag: The etag of this document
    :type etag: unicode
    :param timestamp: The timestamp of this document
    :type timestamp: :class:`datetime.datetime`

    """
    _props_req = [
        'id',
        'content',
        'content_type',
        'etag',
        'timestamp',
    ]

    _props = []

    _props.extend(_props_req)

    def __init__(self, *args, **kwargs):
        self._id = None
        self._content = None
        self._content_type = None
        self._etag = None
        self._timestamp = None

        super(Document, self).__init__(*args, **kwargs)

    @property
    def id(self):
        """The Document id

        :setter: Tries to convert to unicode
        :setter type: str | unicode
        :rtype: unicode

        """
        return self._id

    @id.setter
    def id(self, value):
        if isinstance(value, (bytes, bytearray)) and value is not None:
            value = value.decode("utf-8")
        self._id = value

    @property
    def content_type(self):
        """The Document content type

        :setter: Tries to convert to unicode
        :setter type: str | unicode
        :rtype: unicode

        """
        return self._content_type

    @content_type.setter
    def content_type(self, value):
        if isinstance(value, (bytes, bytearray)) and value is not None:
            value = value.decode("utf-8")
        self._content_type = value

    @property
    def content(self):
        """The Document content

        :setter: Tries to convert to bytearray.
        :setter type: str | unicode | bytearray
        :rtype: bytearray

        """
        return self._content

    @content.setter
    def content(self, value):
        if not isinstance(value, bytearray) and value is not None:
            value = bytearray(value, "utf-8")

        self._content = value

    @property
    def etag(self):
        """The Document etag

        :setter: Tries to convert to unicode
        :setter type: str | unicode
        :rtype: unicode

        """
        return self._etag

    @etag.setter
    def etag(self, value):
        if isinstance(value, (bytes, bytearray)) and value is not None:
            value = value.decode("utf-8")
        self._etag = value

    @property
    def timestamp(self):
        """The Document timestamp.

        :setter: Tries to convert to :class:`datetime.datetime`. If
        no timezone is given, makes a naive `datetime.datetime`.

        Strings will be parsed as ISO 8601 timestamps.

        If a number is provided, it will be interpreted as a UNIX
        timestamp, which by definition is UTC.

        If a `dict` is provided, does `datetime.datetime(**value)`.

        If a `tuple` or a `list` is provided, does
        `datetime.datetime(*value)`. Uses the timezone in the tuple or
        list if provided.

        :setter type: :class:`datetime.datetime` | unicode | str | int | float | dict | tuple | None
        :rtype: :class:`datetime.datetime`

        """
        return self._timestamp

    @timestamp.setter
    def timestamp(self, value):
        if value is None:
            self._timestamp = value
            return

        try:
            self._timestamp = make_datetime(value)
        except TypeError as e:
            message = (
                f"Property 'timestamp' in a 'tincan.{self.__class__.__name__}' "
                f"object must be set with a "
                f"datetime.datetime, str, unicode, int, float, dict "
                f"or None.\n\n{repr(e)}"
            )
            raise TypeError(message) from e



# Copyright 2014 Rustici Software
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from tincan.serializable_base import SerializableBase

"""
.. module:: typed_list
   :synopsis: A wrapper for a list that ensures the list consists of only one type

"""


class TypedList(list, SerializableBase):
    _cls = None

    def __init__(self, *args, **kwargs):
        self._check_cls()
        new_args = [self._make_cls(v) for v in list(*args, **kwargs)]
        super(TypedList, self).__init__(new_args)

    def __setitem__(self, ind, value):
        self._check_cls()
        value = self._make_cls(value)
        super(TypedList, self).__setitem__(ind, value)

    def _check_cls(self):
        """If self._cls is not set, raises ValueError.

        :raises: ValueError
        """
        if self._cls is None:
            raise ValueError("_cls has not been set")

    def _make_cls(self, value):
        """If value is not instance of self._cls, converts and returns
        it. Otherwise, returns value.

        :param value: the thing to make a self._cls from
        :rtype self._cls
        """
        if isinstance(value, self._cls):
            return value
        return self._cls(value)

    def append(self, value):
        self._check_cls()
        value = self._make_cls(value)
        super(TypedList, self).append(value)

    def extend(self, value):
        self._check_cls()
        new_args = [self._make_cls(v) for v in value]
        super(TypedList, self).extend(new_args)

    def insert(self, ind, value):
        self._check_cls()
        value = self._make_cls(value)
        super(TypedList, self).insert(ind, value)



# Copyright 2014 Rustici Software
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from tincan.serializable_base import SerializableBase
from tincan.version import Version
from tincan.extensions import Extensions


class About(SerializableBase):
    """Stores info about this installation of `tincan`.

    :param version: The versions supported. This attribute is required.
    :type version: list of unicode
    :param extensions: Custom user data. This attribute is optional.
    :type extensions: :class:`tincan.Extensions`
    """

    _props_req = [
        'version',
    ]
    _props = [
        'extensions',
    ]
    _props.extend(_props_req)

    def __init__(self, *args, **kwargs):
        self._version = None
        self._extensions = None

        super(About, self).__init__(*args, **kwargs)

    @property
    def version(self):
        """Version for About

        :setter: Sets the version. If None is provided, defaults to
        `[tincan.Version.latest]`. If a string is provided,
        makes a 1-element list containing the string.
        :setter type: list | tuple | str | unicode | None
        :rtype: list
        """
        return self._version

    @version.setter
    def version(self, value):
        def check_version(version):
            """Checks a single version string for validity. Raises
            if invalid.

            :param version: the version string to check
            :type version: list of str or tuple of str or basestring or unicode
            :raises: ValueError
            """
            if version in ['1.0.3', '1.0.2', '1.0.1', '1.0.0', '0.95', '0.9']:
                return

            # Construct the error message
            if isinstance(value, (list, tuple)):
                value_str = repr(version) + ' in ' + repr(value)
            else:
                value_str = repr(version)

            msg = (
                f"Tried to set property 'version' in a 'tincan.{self.__class__.__name__}' object "
                f"with an invalid value: {value_str}\n"
                f"Allowed versions are: {', '.join(map(repr, Version.supported))}"
            )

            raise ValueError(msg)

        if value is None:
            self._version = [Version.latest]
        elif isinstance(value, str):
            check_version(value)
            self._version = [value]
        elif isinstance(value, (list, tuple)):
            for v in value:
                check_version(v)
            self._version = list(value)
        else:
            raise TypeError(
                f"Property 'version' in a 'tincan.{self.__class__.__name__}' object must be set with a "
                f"list, tuple, str, unicode or None. Tried to set it with: {repr(value)}"
            )

    @property
    def extensions(self):
        """Extensions for About

        :setter: Tries to convert to :class:`tincan.Extensions`. If None is provided,
        sets to an empty :class:`tincan.Extensions` dict.
        :setter type: :class:`tincan.Extensions` | dict | None
        :rtype: :class:`tincan.Extensions`
        """
        return self._extensions

    @extensions.setter
    def extensions(self, value):
        if isinstance(value, Extensions):
            self._extensions = value
        elif value is None:
            self._extensions = Extensions()
        else:
            try:
                self._extensions = Extensions(value)
            except Exception as e:
                msg = (
                    f"Property 'extensions' in a 'tincan.{self.__class__.__name__} object must be set with a "
                    f"tincan.Extensions, dict, or None.\n\n"
                )
                msg += repr(e)
                raise TypeError(msg)

    @extensions.deleter
    def extensions(self):
        del self._extensions
